from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    n_attrs = args.ast_graph.nodes[args.n_id]
    var_id = n_attrs["label_field_body"]
    val_id = match_ast_d(args.ast_graph, args.n_id, "for_in_clause") or match_ast_d(
        args.ast_graph,
        args.n_id,
        "if_clause",
    )

    if not val_id:
        val_id = adj_ast(args.ast_graph, args.n_id)[-1]
    return build_assignment_node(args, var_id, val_id, None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    value = node_to_str(graph, args.n_id)
    interpolated_ids = match_ast_group_d(graph, args.n_id, "interpolation")
    declaration_ids = (
        var_id
        for _id in interpolated_ids
        for var_id in adj_ast(graph, _id)
        if graph.nodes[var_id]["label_type"] not in {"{", "}", ",", "interpolation_brace"}
    )

    return build_string_literal_node(args, value, declaration_ids)



from blends.models import (
    NId,
)
from blends.syntax.builders.method_invocation import (
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import node_to_str


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    expr = node_to_str(graph, args.n_id)

    return build_method_invocation_node(args, expr, {})



from blends.models import (
    NId,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    name_id = graph.nodes[args.n_id]["label_field_name"]
    name = node_to_str(graph, name_id)
    body_id = graph.nodes[args.n_id]["label_field_body"]
    return build_class_node(args, name, body_id, None)



from collections.abc import (
    Iterator,
)

from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def search(args: SearchArgs) -> Iterator[SearchResult]:
    if (
        var_id := args.graph.nodes[args.n_id].get("declaration_id")
    ) and args.symbol == args.graph.nodes[var_id].get("variable"):
        yield True, args.n_id

    if (
        var_id
        and args.graph.nodes[var_id]["label_type"] == "Assignment"
        and (val_id := args.graph.nodes[var_id]["variable_id"])
        and args.graph.nodes[val_id].get("symbol") == args.symbol
    ):
        yield True, args.n_id



from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    import_binding_node_attributes,
)
from blends.syntax.builders.import_reference_chain import (
    build_import_reference_chain_to_root,
    extract_module_stem,
)
from blends.syntax.builders.utils import (
    get_next_binding_precedence,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def _bound_import_symbol(*, expression: str, alias: str | None) -> str:
    if isinstance(alias, str) and alias:
        return alias
    parts = expression.split(".")
    return parts[-1] if parts else expression


def build_import_global_node(
    args: SyntaxGraphArgs,
    expression: str,
    module_nodes: set[NId],
    alias: str | None = None,
    module_name: str | None = None,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        expression=expression,
        label_type="Import",
    )

    if alias:
        args.syntax_graph.update_node_attribute(args.n_id, "label_alias", alias)

    for node in module_nodes:
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(node)),
            label_ast="AST",
        )

    if ctx.has_feature_flag("StackGraph"):
        bound = _bound_import_symbol(expression=expression, alias=alias)
        if bound:
            precedence = get_next_binding_precedence(args)
            module_stem = (
                extract_module_stem(module_name)
                if module_name is not None
                else extract_module_stem(expression)
            )
            source_symbol = expression.split(".")[-1] if alias else None
            args.syntax_graph.update_node(
                args.n_id,
                import_binding_node_attributes(
                    symbol=bound,
                    precedence=precedence,
                    import_module_stem=module_stem,
                    import_source_symbol=source_symbol,
                ),
            )
            scope_stack = args.metadata.setdefault("scope_stack", [])
            if scope_stack and (parent_scope := scope_stack[-1]):
                add_edge(
                    args.syntax_graph,
                    Edge(source=parent_scope, sink=args.n_id, precedence=precedence),
                )
            build_import_reference_chain_to_root(args, module_stem, args.n_id)

    return args.n_id



from blends.models import (
    NId,
)
from blends.syntax.builders.method_invocation import (
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    expr_id = graph.nodes[args.n_id]["label_field_function"]
    expr = node_to_str(graph, expr_id)

    arguments_id = graph.nodes[args.n_id]["label_field_arguments"]

    return build_method_invocation_node(
        args, expr, {"expression_id": expr_id, "arguments_id": arguments_id}
    )



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from blends.stack.stacks import (
    ScopeStackNode,
    StackState,
    SymbolStackNode,
)

if TYPE_CHECKING:
    from blends.stack.forward_stitcher import ForwardStitcherConfig
    from blends.stack.partial_path import (
        PartialPath,
        PartialScopeStack,
        PartialSymbolStack,
    )


def _path_priority(path: PartialPath) -> int:
    symbol_depth = len(path.symbol_stack_postcondition.symbols)
    scope_depth = len(path.scope_stack_postcondition.scopes)

    variable_penalty = 0
    if path.symbol_stack_postcondition.variable is not None:
        variable_penalty += 1
    if path.scope_stack_postcondition.variable is not None:
        variable_penalty += 1

    return symbol_depth + scope_depth + variable_penalty


def _should_cancel(config: ForwardStitcherConfig) -> bool:
    if config.is_cancelled is not None and config.is_cancelled():
        return True
    return _deadline_exceeded(config)


def _deadline_exceeded(config: ForwardStitcherConfig) -> bool:
    if config.deadline is None:
        return False
    return time.monotonic() >= config.deadline


def _stack_state_from_partial(
    symbol_stack: PartialSymbolStack, scope_stack: PartialScopeStack
) -> StackState:
    symbol_node = None
    scope_node = None
    if not symbol_stack.can_match_empty():
        symbol_node = SymbolStackNode(symbol_id=0, scopes=None, tail=None)
    if not scope_stack.can_match_empty():
        scope_node = ScopeStackNode(scope_index=0, tail=None)
    return StackState(symbol_stack=symbol_node, scope_stack=scope_node)



from collections.abc import (
    Iterator,
)

from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.models import (
    SyntaxCfgArgs,
)


def iter_with_next(values: list[NId], last: NId | None) -> Iterator[tuple[NId, NId | None]]:
    yield from zip(values, [*values[1:], last], strict=False)


def build(args: SyntaxCfgArgs) -> NId:
    if c_ids := adj_ast(args.graph, args.n_id):
        first_child, *_ = c_ids
        args.graph.add_edge(args.n_id, first_child, label_cfg="CFG")

        for c_id, nxt_id in iter_with_next(list(c_ids), args.nxt_id):
            args.generic(args.fork(c_id, nxt_id))

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.syntax.readers.utils import ensure_nid


def reader(args: SyntaxGraphArgs) -> NId:
    childs = match_ast(args.ast_graph, args.n_id, "if", "else")
    expected_match_count_for_conditional_with_else = 5
    if (
        childs["if"]
        and childs["else"]
        and len(childs) == expected_match_count_for_conditional_with_else
    ):
        return build_if_node(args, ensure_nid(childs["__1__"]), childs["__0__"], childs["__2__"])
    return build_if_node(args, ensure_nid(childs["__0__"]))



from collections.abc import (
    Iterator,
)

from blends.models import (
    NId,
    SyntaxGraph,
)
from blends.symbolic_evaluation.context.search import (
    assignment,
    catch_clause,
    class_body,
    declaration_block,
    file_node,
    for_each,
    for_statement,
    if_statement,
    method_declaration,
    method_invocation,
    try_statement,
    using_statement,
    variable_declaration,
)
from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    Searcher,
    SearchResult,
)
from blends.symbolic_evaluation.models import (
    Path,
)

SEARCHERS: dict[str, Searcher] = {
    "Assignment": assignment.search,
    "CatchClause": catch_clause.search,
    "ClassBody": class_body.search,
    "DeclarationBlock": declaration_block.search,
    "File": file_node.search,
    "ForEachStatement": for_each.search,
    "ForStatement": for_statement.search,
    "If": if_statement.search,
    "MethodDeclaration": method_declaration.search,
    "MethodInvocation": method_invocation.search,
    "UsingStatement": using_statement.search,
    "VariableDeclaration": variable_declaration.search,
    "TryStatement": try_statement.search,
}


def search(
    graph: SyntaxGraph, path: Path, symbol: str, *, def_only: bool
) -> Iterator[SearchResult]:
    for n_id in path:
        if searcher := SEARCHERS.get(graph.nodes[n_id]["label_type"]):
            yield from searcher(SearchArgs(graph, n_id, symbol, def_only))


def search_until_def(graph: SyntaxGraph, path: Path, symbol: str) -> Iterator[NId]:
    for is_def, ref_id in search(graph, path, symbol, def_only=False):
        yield ref_id
        if is_def:
            break


def definition_search(graph: SyntaxGraph, path: Path, symbol: str) -> NId | None:
    for _, ref_id in search(graph, path, symbol, def_only=True):
        return ref_id
    return None



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.switch_section import (
    build_switch_section_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    execution_ids: set[NId] = set()

    for n_id in adj_ast(graph, args.n_id)[2:]:
        if (graph.nodes[n_id].get("label_type") == "expression_statement") and (
            c_ids := adj_ast(graph, n_id)
        ):
            execution_ids.add(c_ids[0])
            continue
        execution_ids.add(n_id)

    return build_switch_section_node(args, "Default", execution_ids)



from blends.models import (
    NId,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    param_node = graph.nodes[args.n_id]
    type_id = param_node.get("label_field_type")
    var_type = param_node.get("label_text")
    return build_parameter_node(
        args=args,
        variable=var_type,
        variable_type=type_id,
        value_id=None,
        c_ids=None,
        modifier=None,
    )



from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    op_attr = args.graph.nodes[args.n_id]
    d_l_expr = generic(args.fork_n_id(op_attr["left_id"])).danger
    d_r_expr = generic(args.fork_n_id(op_attr["right_id"])).danger

    args.evaluation[args.n_id] = d_l_expr or d_r_expr

    if args.method_evaluators and (
        method_evaluator := args.method_evaluators.get("binary_operation")
    ):
        args.evaluation[args.n_id] = method_evaluator(args).danger

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from blends.ctx import ctx
from blends.models import (
    Graph,
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)

_NONE_NAMES: frozenset[str] = frozenset({"None", "none"})


def _name_from_leaf_node(
    graph: Graph,
    type_id: NId,
    node_type: str,
) -> set[str]:
    node = graph.nodes[type_id]
    if node_type == "identifier":
        name = node.get("label_text", "")
        return {name} if name and name not in _NONE_NAMES else set()
    if node_type == "string":
        name = node_to_str(graph, type_id).strip("\"'").split(".")[-1]
        return {name} if name and name not in _NONE_NAMES else set()
    attr_id = node.get("label_field_attribute")
    return _extract_annotation_class_names(graph, attr_id) if attr_id else set()


def _collect_generic_type_names(graph: Graph, type_id: NId) -> set[str]:
    type_param_id = match_ast_d(graph, type_id, "type_parameter")
    if not type_param_id:
        return set()
    names: set[str] = set()
    for child_id in adj_ast(graph, type_param_id):
        if graph.nodes.get(child_id, {}).get("label_type", "") in {",", "[", "]"}:
            continue
        names |= _extract_annotation_class_names(graph, child_id)
    return names - _NONE_NAMES


def _collect_union_names(graph: Graph, type_id: NId) -> set[str]:
    names: set[str] = set()
    for child_id in adj_ast(graph, type_id):
        if graph.nodes.get(child_id, {}).get("label_type", "") == "|":
            continue
        names |= _extract_annotation_class_names(graph, child_id)
    return names - _NONE_NAMES


_LEAF_NODE_TYPES: frozenset[str] = frozenset({"identifier", "string", "attribute"})
_UNION_NODE_TYPES: frozenset[str] = frozenset({"binary_operator", "union_type"})


def _extract_annotation_class_names(graph: Graph, type_id: NId) -> set[str]:
    node: dict[str, str] = graph.nodes.get(type_id, {})
    node_type = node.get("label_type", "")

    if node_type == "type":
        children = adj_ast(graph, type_id)
        return _extract_annotation_class_names(graph, children[0]) if children else set()

    if node_type in _LEAF_NODE_TYPES:
        return _name_from_leaf_node(graph, type_id, node_type)

    if node_type == "generic_type":
        return _collect_generic_type_names(graph, type_id)

    if node_type in _UNION_NODE_TYPES:
        return _collect_union_names(graph, type_id)

    return set()


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    var_type = None
    var_name = "UnnamedVar"
    annotation_class_names: set[str] | None = None

    type_id = n_attrs.get("label_field_type")
    if type_id:
        var_type = node_to_str(graph, type_id)
        if ctx.has_feature_flag("StackGraph"):
            annotation_class_names = _extract_annotation_class_names(graph, type_id)

    identifier_id = n_attrs.get("label_field_name") or match_ast_d(graph, args.n_id, "identifier")
    if identifier_id:
        var_name = node_to_str(graph, identifier_id)

    value_id = n_attrs.get("label_field_value")
    return build_parameter_node(
        args=args,
        variable=var_name,
        variable_type=var_type,
        value_id=value_id,
        annotation_class_names=annotation_class_names,
    )



from blends.models import (
    NId,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.builders.literal import build_literal_node
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]

    condition_id = n_attrs.get("label_field_condition")
    if not condition_id:
        return build_literal_node(args, "unless", "unless")

    false_id = n_attrs.get("label_field_body")

    return build_if_node(args, condition_id, None, false_id)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.while_statement import (
    build_while_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    condition_id = n_attrs["label_field_condition"]
    block = n_attrs["label_field_body"]

    if graph.nodes[condition_id].get("label_type") == "parenthesized_expression":
        c_ids = match_ast(graph, condition_id)
        condition_id = c_ids.get("__1__")

    return build_while_statement_node(args, block, condition_id)



from collections.abc import Callable, Iterator
from typing import TypeVar
from typing_extensions import ParamSpec

_T = TypeVar("_T")
_P = ParamSpec("_P")
_E = TypeVar("_E", bound=BaseException)

class MarkDecorator:
    def __call__(self, func: Callable[_P, _T]) -> Callable[_P, _T]: ...

class MarkGenerator:
    def blends_test_group(self, value: str) -> MarkDecorator: ...

mark: MarkGenerator

class Parser:
    def addoption(
        self,
        name: str,
        *,
        action: str | None = None,
        metavar: str | None = None,
        **kwargs: object,
    ) -> None: ...

class Config:
    def getoption(self, name: str, default: str | None = None) -> str | None: ...

class Mark:
    args: tuple[str, ...]

class Item:
    config: Config
    def iter_markers(self, name: str | None = None) -> Iterator[Mark]: ...

class Module(Item): ...

def skip(reason: str) -> None: ...

class RaisesContext:
    def __enter__(self) -> None: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> bool: ...

def raises(
    expected_exception: type[_E] | tuple[type[_E], ...],
    *,
    match: str | None = None,
) -> RaisesContext: ...



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from blends.stack.bridge_paths import index_bridge_paths_for_file
from blends.stack.criteria import (
    is_definition_endpoint,
    is_reference_endpoint,
)
from blends.stack.multi_file_stitcher import MultiFileForwardStitcher
from blends.stack.partial_path_db import PartialPathDb
from blends.stack.selection import edge_list_shadows

if TYPE_CHECKING:
    from blends.stack.forward_stitcher import ForwardStitcherConfig
    from blends.stack.partial_path import (
        FileHandle,
        PartialPath,
        PartialPathDatabase,
        PartialSymbolStack,
    )
    from blends.stack.selection import PartialPathEdge
    from blends.stack.view import StackGraphView


@dataclass(frozen=True, slots=True)
class MultiFileResolutionStats:
    phases_executed: int
    candidates_considered: int
    concatenations_succeeded: int
    complete_paths_found: int
    cancelled: bool
    cycle_guard_hits: int
    queue_pruned_count: int
    files_considered: int


@dataclass(frozen=True, slots=True)
class MultiFileResolutionRequest:
    file_handle: FileHandle
    ref_node_index: int
    config: ForwardStitcherConfig | None = None
    include_stats: bool = False


@dataclass(frozen=True, slots=True)
class _MultiFileDefinitionCandidate:
    definition_node_index: int
    file_handle: FileHandle
    edges: tuple[PartialPathEdge, ...]


def _build_source_only_db(
    database: PartialPathDatabase,
    view: StackGraphView,
    source_fh: FileHandle,
) -> PartialPathDb:
    db = PartialPathDb(symbols=view.symbols)
    record = database.get_file(source_fh)
    if record is not None:
        db.add_file(record)
    return db


def _build_stem_index(database: PartialPathDatabase) -> dict[str, FileHandle]:
    return {
        Path(record.file_path).stem: fh
        for fh in database.list_files()
        if (record := database.get_file(fh)) is not None
    }


def _extract_module_stem(
    symbol_stack: PartialSymbolStack,
    db: PartialPathDb,
) -> str | None:
    if not symbol_stack.symbols:
        return None
    return db.symbols.id_to_symbol[symbol_stack.symbols[0].symbol_id]


def resolve_definitions_multi_file(
    view: StackGraphView,
    database: PartialPathDatabase,
    *,
    request: MultiFileResolutionRequest,
    views: dict[FileHandle, StackGraphView] | None = None,
) -> list[tuple[FileHandle, int]] | tuple[list[tuple[FileHandle, int]], MultiFileResolutionStats]:
    if not is_reference_endpoint(view, request.ref_node_index):
        return _empty_multi_result(include_stats=request.include_stats)

    if request.file_handle not in database.list_files():
        return _empty_multi_result(include_stats=request.include_stats)
    if not _ref_record_matches_view(view, request, database):
        return _empty_multi_result(include_stats=request.include_stats)

    db = _build_source_only_db(database, view, request.file_handle)
    all_views: dict[FileHandle, StackGraphView] = (
        views if views is not None else {request.file_handle: view}
    )
    index_bridge_paths_for_file(view, file_handle=request.file_handle, db=db)

    stitcher = MultiFileForwardStitcher.from_references(
        view=view,
        file_handle=request.file_handle,
        db=db,
        reference_nodes=(request.ref_node_index,),
        config=request.config,
        views=all_views,
    )

    pairs, latest_stats, files_loaded = _run_resolution_multi_file(
        view, stitcher, database, all_views, db
    )
    return _finalize_multi_result(
        pairs, latest_stats, files_loaded, include_stats=request.include_stats
    )


def _empty_multi_result(
    *,
    include_stats: bool,
) -> list[tuple[FileHandle, int]] | tuple[list[tuple[FileHandle, int]], MultiFileResolutionStats]:
    if include_stats:
        return [], MultiFileResolutionStats(
            phases_executed=0,
            candidates_considered=0,
            concatenations_succeeded=0,
            complete_paths_found=0,
            cancelled=False,
            cycle_guard_hits=0,
            queue_pruned_count=0,
            files_considered=0,
        )
    return []


def _collect_multi_file_candidates(
    views: dict[FileHandle, StackGraphView],
    complete_paths: tuple[tuple[FileHandle, PartialPath], ...],
    ref_view: StackGraphView,
) -> list[_MultiFileDefinitionCandidate]:
    new_candidates: list[_MultiFileDefinitionCandidate] = []
    for def_fh, path in complete_paths:
        def_view = views.get(def_fh, ref_view)
        if not is_definition_endpoint(def_view, path.end_node_index):
            continue
        new_candidates.append(
            _MultiFileDefinitionCandidate(
                definition_node_index=path.end_node_index,
                file_handle=def_fh,
                edges=path.edges,
            )
        )
    return new_candidates


def _prune_multi_shadowed(
    candidates: list[_MultiFileDefinitionCandidate],
) -> list[_MultiFileDefinitionCandidate]:
    kept: list[_MultiFileDefinitionCandidate] = []
    for candidate in sorted(candidates, key=lambda c: (c.file_handle, c.definition_node_index)):
        is_dominated = any(edge_list_shadows(k.edges, candidate.edges) for k in kept)
        if is_dominated:
            continue
        kept = [k for k in kept if not edge_list_shadows(candidate.edges, k.edges)]
        kept.append(candidate)
    return kept


def _sort_multi_deterministically(
    candidates: list[_MultiFileDefinitionCandidate],
) -> list[_MultiFileDefinitionCandidate]:
    def sort_key(
        candidate: _MultiFileDefinitionCandidate,
    ) -> tuple[tuple[int, ...], tuple[int, ...], int, int]:
        edge_sources = tuple(edge.source_node_index for edge in candidate.edges)
        edge_precedences = tuple(-edge.precedence for edge in candidate.edges)
        return (
            edge_sources,
            edge_precedences,
            candidate.file_handle,
            candidate.definition_node_index,
        )

    return sorted(candidates, key=sort_key)


def _select_multi_file_definitions(
    candidates: list[_MultiFileDefinitionCandidate],
) -> list[tuple[FileHandle, int]]:
    if not candidates:
        return []
    pruned = _prune_multi_shadowed(candidates)
    ordered = _sort_multi_deterministically(pruned)
    seen: set[tuple[int, int]] = set()
    pairs: list[tuple[FileHandle, int]] = []
    for candidate in ordered:
        key = (candidate.file_handle, candidate.definition_node_index)
        if key in seen:
            continue
        seen.add(key)
        pairs.append(key)
    return pairs


def _run_resolution_multi_file(
    view: StackGraphView,
    stitcher: MultiFileForwardStitcher,
    database: PartialPathDatabase,
    all_views: dict[FileHandle, StackGraphView],
    db: PartialPathDb,
) -> tuple[list[tuple[FileHandle, int]], MultiFileResolutionStats | None, int]:
    candidates: list[_MultiFileDefinitionCandidate] = []
    loaded: set[FileHandle] = {stitcher.file_handle}
    stem_index = _build_stem_index(database)
    result = None

    while not stitcher.is_complete():
        result = stitcher.process_next_phase()
        for _fh, sym_stack in result.load_requests.root_requests:
            if (stem := _extract_module_stem(sym_stack, db)) is None:
                continue
            if (target_fh := stem_index.get(stem)) is None or target_fh in loaded:
                continue
            loaded.add(target_fh)
            if (rec := database.get_file(target_fh)) is not None:
                db.add_file(rec)
            if (fh_view := all_views.get(target_fh)) is not None:
                index_bridge_paths_for_file(fh_view, file_handle=target_fh, db=db)
                stitcher.views[target_fh] = fh_view
        candidates.extend(_collect_multi_file_candidates(all_views, result.complete_paths, view))

    pairs = _select_multi_file_definitions(candidates)
    if result is None:
        return pairs, None, len(loaded)
    latest_stats = MultiFileResolutionStats(
        phases_executed=result.stats.phases_executed,
        candidates_considered=result.stats.candidates_considered,
        concatenations_succeeded=result.stats.concatenations_succeeded,
        complete_paths_found=result.stats.complete_paths_found,
        cancelled=result.stats.cancelled,
        cycle_guard_hits=result.stats.cycle_guard_hits,
        queue_pruned_count=result.stats.queue_pruned_count,
        files_considered=len(loaded),
    )
    return pairs, latest_stats, len(loaded)


def _finalize_multi_result(
    pairs: list[tuple[FileHandle, int]],
    latest_stats: MultiFileResolutionStats | None,
    files_considered: int,
    *,
    include_stats: bool,
) -> list[tuple[FileHandle, int]] | tuple[list[tuple[FileHandle, int]], MultiFileResolutionStats]:
    if include_stats:
        if latest_stats is None:
            latest_stats = MultiFileResolutionStats(
                phases_executed=0,
                candidates_considered=0,
                concatenations_succeeded=0,
                complete_paths_found=0,
                cancelled=False,
                cycle_guard_hits=0,
                queue_pruned_count=0,
                files_considered=files_considered,
            )
        return pairs, latest_stats
    return pairs


def _ref_record_matches_view(
    view: StackGraphView,
    request: MultiFileResolutionRequest,
    database: PartialPathDatabase,
) -> bool:
    record = database.get_file(request.file_handle)
    if record is None:
        return False
    if record.file_path and view.file_path:
        return record.file_path == view.file_path
    return True



from blends.models import NId
from blends.query import adj_ast
from blends.syntax.builders.object import build_object_node
from blends.syntax.models import SyntaxGraphArgs


def reader(args: SyntaxGraphArgs) -> NId:
    ignore_types = {"[", "]", ":", ","}
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)
    return build_object_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"] not in ignore_types),
    )



import re

from blends.models import (
    NId,
    SyntaxGraph,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
)

# https://learn.microsoft.com/en-us/dotnet/api/system.web.httprequest
NET_SYSTEM_HTTP_METHODS = {
    "Cookies",
    "Form",
    "Headers",
    "Params",
    "Path",
    "PathInfo",
    "QueryString",
    "ReadLine",
    "GetString",
}
NET_CONNECTION_OBJECTS = {
    "TcpClient",
    "TcpListener",
    "SqlCommand",
    "WebClient",
}

JAVA_CONNECTION_METHODS = {
    "batchUpdate",
    "getBytes",
    "getHeader",
    "getHeaderNames",
    "getHeaders",
    "getName",
    "getParameter",
    "getTheParameter",
    "getParameterMap",
    "getParameterNames",
    "getParameterValues",
    "getQueryString",
    "getValue",
    "query",
    "queryForList",
    "queryForMap",
    "queryForObject",
    "queryForRowSet",
    "getInputStream",
    "getReader",
    "getCookies",
}

ANGULAR_USER_INPUTS = [
    r"\bthis\.route\.params\b",
    r"\bthis\.route\.queryParams\b",
    r"\bthis\.route\.fragment\b",
    r"\bthis\.http\.post",
    r"\bthis\.http\.put",
    r"\bthis\.http\.patch",
    r"\bdocument\.cookie\b",
    r"\blocalStorage\.getItem\(",
    r"\bsessionStorage\.getItem\(",
    r"\bwindow\.location\b",
    r"\bwindow\.name\b",
    r"\bwindow\.history\b",
    r"\bdocument\.URL\b",
    r"\bdocument\.location\b",
    r"\bdocument\.referrer\b",
]

JS_TS_HTTP_INPUTS = {
    r"\b(req|request)\.body\b",
    r"\b(req|request)\.params\b",
    r"\b(req|request)\.query\b",
    r"\b(req|request)\.originalUrl\b",
    r"\b(req|request)\.baseUrl\b",
    r"\b(req|request)\.path\b",
    r"\b(req|request)\.headers\b",
    r"\b(req|request)\.cookies\b",
    r"\b(req|request)\.file\b",
}

PYTHON_INPUTS = {
    "request.COOKIES.get",
    "request.GET",
    "request.GET.get",
    "request.POST.get",
    "request.args",
    "request.args.get",
    "request.data.get",
    "request.files",
    "request.files.get",
    "request.headers",
    "request.headers.get",
    "request.META.get",
    "request.form.get",
    "request.values.get",
    "request.json.get",
    "request.get_json",
    "request.args.to_dict",
    "request.get_data",
}

PHP_INPUTS = {"_GET", "_POST", "_REQUEST", "_FILES", "_COOKIE"}

PYTHON_REQUESTS_LIBRARIES = {
    "requests",
    "urllib",
    "urllib2",
    "urllib3",
    "httplib2",
    "httplib",
    "http",
    "treq",
    "aiohttp",
}
SCALA_INPUTS = {
    "request.uri.query.params.get",
    "request.uri",
    "request.headers.get",
    "request.cookies.get",
    "request.as[String]",
    "request.decode",
    "request.attributes.get",
    "request.getQueryString",
    "request.queryString",
    "request.body.asText",
    "request.body.asFormUrlEncoded",
    "request.body",
    "request.path",
}

SCALA_HTTP_LIBRARIES = {
    "org.http4s.client",
    "sttp.client3",
    "akka.http.scaladsl.Http",
    "play.api.libs.ws",
    "scalaj.http",
    "dispatch._",
    "feign",
}

RUBY_INPUTS = {
    "params",
    "request.query_parameters",
    "request.request_parameters",
    "cookies",
    "request.headers",
}

KOTLIN_INPUTS = {
    "call.parameters",
    "call.request.queryParameters",
    "call.request.headers",
    "call.request.cookies",
    "call.receiveText",
    "call.receiveParameters",
    "request.getParameter",
    "request.getHeader",
    "request.queryString",
    "request.cookies",
}

INSECURE_ALGOS = {
    "none",
    "blowfish",
    "bf",
    "des",
    "desede",
    "rc2",
    "rc4",
    "rsa",
    "3des",
}
INSECURE_MODES = {"cbc", "ecb", "ofb"}
INSECURE_HASHES = {"md2", "md4", "md5", "sha1", "sha-1"}

CREDENTIAL_SETTINGS = {
    "EMAIL_HOST_PASSWORD",
    "API_KEY",
    "SECRET_KEY",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_S3_SECRET_ACCESS_KEY",
    "SOCIAL_AUTH_GOOGLE_OAUTH_SECRET",
    "SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET",
    "SOCIAL_AUTH_GOOGLE_PLUS_SECRET",
    "AUTH_LDAP_BIND_PASSWORD",
    "SECRET_KEY_FALLBACKS",
}

SWIFT_INPUTS = (
    "req.content",
    "req.query",
    "req.parameters",
    "req.body",
    "req.cookies",
    "req.headers",
    "req.method",
    "req.url",
)

SPRING_INPUTS = (
    "RequestParam",
    "RequestHeader",
    "CookieValue",
    "RequestBody",
    "PathVariable",
    "MatrixVariable",
    "ModelAttribute",
)


def concatenate_name(graph: SyntaxGraph, n_id: NId, name: str | None = None) -> str:
    prev_str = ""
    if name:
        prev_str = name if name.startswith("::") else f".{name}"

    node_type = graph.nodes[n_id]["label_type"]
    if node_type == "MethodInvocation":
        expr = graph.nodes[n_id]["expression"]
        if graph.nodes[n_id].get("object_id") and (next_node := match_ast(graph, n_id)["__0__"]):
            expr = concatenate_name(graph, next_node, expr)
    elif node_type == "SymbolLookup":
        expr = graph.nodes[n_id]["symbol"]
    else:
        expr = ""
    return str(expr + prev_str)


def check_swift_inputs(graph: SyntaxGraph, n_id: NId) -> bool:
    node_id = n_id
    if graph.nodes[node_id]["label_type"] == "If" and (
        block_id := graph.nodes[node_id]["condition_id"]
    ):
        node_id = block_id
    if graph.nodes[node_id]["label_type"] == "TryStatement" and (
        block_id := graph.nodes[node_id]["block_id"]
    ):
        node_id = block_id
    return bool(
        (
            graph.nodes[node_id]["label_type"] == "MethodInvocation"
            or graph.nodes[node_id]["label_type"] == "MemberAccess"
        )
        and (expr := graph.nodes[node_id].get("expression"))
        and expr.startswith(SWIFT_INPUTS)
    )


def owasp_user_connection(args: SymbolicEvalArgs) -> bool:
    ma_attr = args.graph.nodes[args.n_id]
    return bool(
        ma_attr["expression"] in JAVA_CONNECTION_METHODS
        and (obj_id := ma_attr.get("object_id"))
        and generic(args.fork(n_id=obj_id, evaluation={}, triggers=set())).triggers
        == {"userconnection"},
    )


def check_js_ts_http_inputs(args: SymbolicEvalArgs) -> bool:
    n_attrs = args.graph.nodes[args.n_id]
    member_access = f"{n_attrs['expression']}.{n_attrs['member']}"
    return next(
        (True for expr in JS_TS_HTTP_INPUTS if re.match(expr, member_access) is not None),
        False,
    )


def check_python_inputs(args: SymbolicEvalArgs) -> bool:
    n_attrs = args.graph.nodes[args.n_id]
    if n_attrs["label_type"] != "MemberAccess":
        return False
    member_access = f"{n_attrs['expression']}.{n_attrs['member']}"
    return member_access in PYTHON_INPUTS


def check_php_inputs(graph: SyntaxGraph, n_id: NId) -> bool:
    node = graph.nodes[n_id]
    if node["label_type"] != "ElementAccess":
        return False
    return graph.nodes[node["expression_id"]].get("symbol", "") in PHP_INPUTS


def check_kotlin_inputs(graph: SyntaxGraph, n_id: NId) -> bool:
    node = graph.nodes[n_id]
    if not (expr_id := node.get("expression_id")):
        return False
    expr_node = graph.nodes[expr_id]
    if expr_node.get("label_type") != "MemberAccess":
        return False
    member_access = f"{expr_node.get('expression')}.{expr_node.get('member')}"
    return member_access in KOTLIN_INPUTS


def check_ruby_inputs(graph: SyntaxGraph, n_id: NId) -> bool:
    node = graph.nodes[n_id]
    return bool(
        (node["label_type"] == "ElementAccess")
        and (expr_id := node.get("expression_id"))
        and (expr_node := graph.nodes[expr_id])
        and (
            (
                (expr_node["label_type"] != "MethodInvocation")
                and (expr_node["symbol"] in RUBY_INPUTS)
            )
            or (
                (expr_node["label_type"] == "MethodInvocation")
                and (obj_id := expr_node.get("object_id"))
                and (obj_node := graph.nodes[obj_id])
                and (expr_id := expr_node.get("expression_id"))
                and (expr_node := graph.nodes[expr_id])
                and (member_access := f"{obj_node['symbol']}.{expr_node.get('symbol')}")
                and (member_access in RUBY_INPUTS)
            )
        )
    )


def check_scala_inputs(graph: SyntaxGraph, n_id: NId) -> bool:
    return bool(
        (method_name := concatenate_name(graph, n_id))
        and any(
            method_name.startswith(input_pattern) or method_name in input_pattern
            for input_pattern in SCALA_INPUTS
            if input_pattern
        )
    )


def check_spring_scala_inputs(graph: SyntaxGraph, n_id: NId) -> bool:
    if graph.nodes[n_id]["label_type"] == "Parameter" and (param_data := adj_ast(graph, n_id)):
        for param_id in param_data:
            if (
                graph.nodes[param_id]["label_type"] == "Annotation"
                and (name := graph.nodes[param_id].get("name"))
                and (name in SPRING_INPUTS)
            ):
                return True
    return False



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.try_statement import (
    build_try_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    block_id = match_ast_d(graph, args.n_id, "block")
    if not block_id:
        block_id = adj_ast(graph, args.n_id)[-1]

    catch_blocks = match_ast_group_d(args.ast_graph, args.n_id, "catch_block")
    finally_block = match_ast_d(graph, args.n_id, "label_field_finally_block")

    return build_try_statement_node(args, block_id, catch_blocks, finally_block, None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.file import (
    build_file_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    body_id = match_ast_d(args.ast_graph, args.n_id, "body")
    c_ids = adj_ast(args.ast_graph, body_id) if body_id else adj_ast(args.ast_graph, args.n_id)
    return build_file_node(args, iter(c_ids))



from collections.abc import (
    Iterator,
)

from blends.query import (
    adj_ast,
)
from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def search(args: SearchArgs) -> Iterator[SearchResult]:
    if args.graph.nodes[args.n_id].get("name") == args.symbol:
        yield True, args.n_id
    else:
        pl_id = args.graph.nodes[args.n_id].get("parameters_id") or args.n_id
        if args.graph.nodes[pl_id].get("label_type") == "Parameter":
            if args.symbol == args.graph.nodes[pl_id].get("variable"):
                yield True, pl_id
        else:
            params_ids = adj_ast(args.graph, pl_id, label_type="Parameter")
            for c_id in params_ids:
                if args.symbol == args.graph.nodes[c_id].get("variable"):
                    yield True, c_id
                    break



from collections.abc import Callable
from pathlib import Path

import pytest

from blends import get_graphs_from_path
from blends.models import NId, SyntaxGraph
from blends.query import matching_nodes
from blends.symbolic_evaluation.evaluate import evaluate
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)
from blends.symbolic_evaluation.utils import get_backward_paths

TEST_DATA_DIR = Path(__file__).parent.parent / "data" / "test_files" / "symbolic_evaluation"


def make_trigger_evaluator(node_type: str) -> Callable[[SymbolicEvalArgs], SymbolicEvaluation]:
    def evaluator(args: SymbolicEvalArgs) -> SymbolicEvaluation:
        args.triggers.add(node_type)
        return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)

    return evaluator


def make_danger_evaluator(node_type: str) -> Callable[[SymbolicEvalArgs], SymbolicEvaluation]:
    def evaluator(args: SymbolicEvalArgs) -> SymbolicEvaluation:
        args.triggers.add(node_type)
        args.evaluation[args.n_id] = True
        return SymbolicEvaluation(danger=True, triggers=args.triggers)

    return evaluator


def build_method_evaluators(
    node_types: list[str],
) -> dict[str, Callable[[SymbolicEvalArgs], SymbolicEvaluation]]:
    return {node_type: make_trigger_evaluator(node_type) for node_type in node_types}


def get_syntax_graph(file_path: Path) -> SyntaxGraph:
    graphs = get_graphs_from_path(file_path)
    assert graphs.syntax_graph is not None, f"Failed to parse {file_path}"
    return graphs.syntax_graph


def evaluate_node(
    graph: SyntaxGraph,
    n_id: NId,
    method_evaluators: dict[str, Callable[[SymbolicEvalArgs], SymbolicEvaluation]] | None = None,
) -> SymbolicEvaluation | None:
    for path in get_backward_paths(graph, n_id):
        result = evaluate(graph, path, n_id, None, method_evaluators)
        if result is not None:
            return result
    return None


@pytest.mark.blends_test_group("functional")
class TestSymbolicEvaluationFunctional:
    def test_python_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "python.py")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        node_types = ["method_invocation", "member_access", "symbol_lookup"]
        evaluators = build_method_evaluators(node_types)
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers
        assert "member_access" in result.triggers
        assert "symbol_lookup" in result.triggers

    def test_python_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "python.py")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = {
            "parameter": make_danger_evaluator("parameter"),
            "method_invocation": make_trigger_evaluator("method_invocation"),
        }
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "parameter" in result.triggers

    def test_python_literal_not_dangerous(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "python.py")

        literals = matching_nodes(graph, label_type="Literal")
        assert literals, "No Literal nodes found"

        evaluators = build_method_evaluators(["literal"])
        result = evaluate_node(graph, literals[0], evaluators)

        assert result is not None
        assert result.danger is False
        assert "literal" in result.triggers

    def test_python_binary_operation_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "python.py")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers
        assert "symbol_lookup" in result.triggers

    def test_javascript_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "javascript.js")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_javascript_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "javascript.js")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers

    def test_java_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "java.java")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_java_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "java.java")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers

    def test_typescript_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "typescript.ts")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_typescript_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "typescript.ts")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers

    def test_go_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "go.go")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_go_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "go.go")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers

    def test_kotlin_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "kotlin.kt")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_kotlin_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "kotlin.kt")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers

    def test_csharp_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "c_sharp.cs")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_csharp_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "c_sharp.cs")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers

    def test_php_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "php.php")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_php_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "php.php")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers

    def test_scala_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "scala.scala")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_scala_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "scala.scala")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers

    def test_swift_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "swift.swift")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_swift_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "swift.swift")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers

    def test_dart_method_invocation_evaluation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "dart.dart")

        method_invocations = matching_nodes(graph, label_type="MethodInvocation")
        assert method_invocations, "No MethodInvocation nodes found"

        evaluators = build_method_evaluators(["method_invocation", "member_access"])
        result = evaluate_node(graph, method_invocations[0], evaluators)

        assert result is not None
        assert "method_invocation" in result.triggers

    def test_dart_danger_propagation(self) -> None:
        graph = get_syntax_graph(TEST_DATA_DIR / "dart.dart")

        binary_ops = matching_nodes(graph, label_type="BinaryOperation")
        assert binary_ops, "No BinaryOperation nodes found"

        evaluators = {
            "binary_operation": make_trigger_evaluator("binary_operation"),
            "symbol_lookup": make_danger_evaluator("symbol_lookup"),
        }
        result = evaluate_node(graph, binary_ops[0], evaluators)

        assert result is not None
        assert result.danger is True
        assert "binary_operation" in result.triggers



from __future__ import annotations

from typing import TYPE_CHECKING

from blends.stack.node_kinds import (
    StackGraphNodeKind,
)

if TYPE_CHECKING:
    from blends.stack.stacks import (
        StackState,
    )
    from blends.stack.view import (
        StackGraphView,
    )


def is_reference_endpoint(view: StackGraphView, node_index: int) -> bool:
    kind = view.kind_at(node_index)
    is_push = kind in {
        StackGraphNodeKind.PUSH_SYMBOL.value,
        StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value,
    }
    if not is_push:
        return False
    if node_index in {0, 1}:
        return False
    return view.node_is_reference[node_index]


def is_definition_endpoint(view: StackGraphView, node_index: int) -> bool:
    kind = view.kind_at(node_index)
    is_pop = kind in {
        StackGraphNodeKind.POP_SYMBOL.value,
        StackGraphNodeKind.POP_SCOPED_SYMBOL.value,
    }
    if not is_pop:
        return False
    if node_index in {0, 1}:
        return False
    return view.node_is_definition[node_index]


def is_exported_scope_endpoint(view: StackGraphView, node_index: int) -> bool:
    return (
        view.kind_at(node_index) == StackGraphNodeKind.SCOPE.value
        and node_index in view.exported_scopes
    )


def is_root_endpoint(view: StackGraphView, node_index: int) -> bool:
    return view.kind_at(node_index) == StackGraphNodeKind.ROOT.value


def is_jump_to_boundary(view: StackGraphView, node_index: int) -> bool:
    return view.kind_at(node_index) == StackGraphNodeKind.JUMP_TO.value


def is_endpoint(view: StackGraphView, node_index: int) -> bool:
    return (
        is_reference_endpoint(view, node_index)
        or is_definition_endpoint(view, node_index)
        or is_exported_scope_endpoint(view, node_index)
        or is_root_endpoint(view, node_index)
    )


def is_import_binding_endpoint(view: StackGraphView, node_index: int) -> bool:
    return is_definition_endpoint(view, node_index) and view.node_is_import_binding[node_index]


def is_complete_definition_path(
    view: StackGraphView,
    *,
    start_node_index: int,
    end_node_index: int,
    start_state: StackState,
    end_state: StackState,
) -> bool:
    if not is_reference_endpoint(view, start_node_index):
        return False
    if not is_definition_endpoint(view, end_node_index):
        return False
    if start_state.symbol_stack is not None:
        return False
    if start_state.scope_stack is not None:
        return False
    if end_state.symbol_stack is not None:
        return False
    return end_state.scope_stack is None



from blends.models import (
    NId,
)
from blends.syntax.builders.catch_clause import (
    build_catch_clause_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    catch_node = args.ast_graph.nodes[args.n_id]
    block_node = catch_node["label_field_body"]
    param_id = catch_node.get("label_field_parameter")
    return build_catch_clause_node(args, block_node, param_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.export_statement import (
    build_export_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expression = None
    export_block = graph.nodes[args.n_id].get("label_field_declaration")

    if export_block and graph.nodes[export_block]["label_type"] in {
        "identifier",
        "string",
    }:
        expression = node_to_str(args.ast_graph, export_block)
        export_block = None

    return build_export_statement_node(args, expression, export_block)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    var_id = graph.nodes[args.n_id]["label_field_name"]
    var_name = node_to_str(graph, var_id)
    var_type = match_ast_d(graph, args.n_id, "type_identifier")

    return build_variable_declaration_node(args, var_name, var_type, var_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.object import (
    build_object_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    match_identifier = graph.nodes[args.n_id]["label_field_name"]
    property_name = graph.nodes[match_identifier].get("label_text")

    accessors = adj_ast(graph, args.n_id, label_type="accessor_declaration", depth=2)

    return build_object_node(args, iter(accessors), property_name)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]

    condition_id = n_attrs["label_field_condition"]
    if graph.nodes[condition_id]["label_type"] == "parenthesized_expression":
        condition_id = match_ast(graph, condition_id).get("__1__")

    true_id = n_attrs["label_field_consequence"]
    if graph.nodes[true_id]["label_type"] == "expression_statement":
        true_id = match_ast(graph, true_id)["__0__"]

    false_id = n_attrs.get("label_field_alternative")
    return build_if_node(args, condition_id, true_id, false_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    symbol_n_id = args.n_id

    if graph.nodes[args.n_id].get("label_type") == "name" and (c_ids := adj_ast(graph, args.n_id)):
        symbol_n_id = c_ids[0]

    symbol = args.ast_graph.nodes[symbol_n_id]["label_text"]

    return build_symbol_lookup_node(args, symbol)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_group_d,
)
from blends.syntax.builders.reserved_word import (
    build_reserved_word_node,
)
from blends.syntax.builders.try_statement import (
    build_try_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs = match_ast(graph, args.n_id, "call", "assignment")
    block_node = childs.get("assignment")
    if not block_node:
        direct_childs = adj_ast(graph, args.n_id)
        min_children_for_begin_block = 2
        if len(direct_childs) < min_children_for_begin_block:
            return build_reserved_word_node(args, "begin")
        block_node = direct_childs[1]

    try_block_id = childs.get("call")
    catch_blocks = match_ast_group_d(graph, args.n_id, "rescue")
    return build_try_statement_node(args, block_node, catch_blocks, try_block_id, None)



from blends.models import (
    NId,
)
from blends.query import adj_ast, match_ast_d
from blends.syntax.builders.element_access import (
    build_element_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expr_id = match_ast_d(graph, args.n_id, "navigation_expression")
    if not expr_id:
        expr_id = adj_ast(graph, args.n_id)[0]

    arguments_id = match_ast_d(graph, args.n_id, "string_literal")
    return build_element_access_node(args, expr_id, arguments_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.parenthesized_expression import (
    build_parenthesized_expression_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    childs = adj_ast(args.ast_graph, args.n_id)
    expected_children_for_parenthesized_expression = 3
    c_id = (
        childs[1] if len(childs) == expected_children_for_parenthesized_expression else childs[-2]
    )
    return build_parenthesized_expression_node(args, c_id)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
    match_ast_group,
    match_ast_group_d,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    ListChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    name_id = graph.nodes[args.n_id].get("label_field_name")
    name = node_to_str(graph, name_id) if name_id else None

    block_id = match_ast_d(graph, args.n_id, "function_body")
    if not block_id:
        block_id = match_ast_d(graph, args.n_id, "statements")

    parameters_id = graph.nodes[args.n_id].get("label_field_parameters") or match_ast_d(
        graph, args.n_id, "function_value_parameters"
    )

    modifiers_ids = match_ast_group_d(graph, args.n_id, "modifiers")
    if modifiers_ids:
        first_modifier_id = modifiers_ids[0]
        annotation_ids_raw = match_ast_group(
            graph,
            first_modifier_id,
            "annotation",
        )
        annotation_ids = annotation_ids_raw.get("annotation", [])
        if len(annotation_ids) == 0:
            modifiers_ids = []

    direct: DirectChildren = {}
    if block_id:
        direct["block_id"] = block_id
    if parameters_id:
        direct["parameters_id"] = NId(parameters_id)

    list_children: ListChildren = {}
    if modifiers_ids:
        list_children["modifiers_ids"] = modifiers_ids

    return build_method_declaration_node(args, name, direct, list_children)



from collections.abc import (
    Callable,
)
from typing import (
    NamedTuple,
)

from blends.models import (
    NId,
    SyntaxGraph,
)
from blends.symbolic_evaluation.context.search import (
    definition_search,
)
from blends.symbolic_evaluation.models import (
    Path,
)
from blends.symbolic_evaluation.utils import (
    get_lookup_path,
)


class SolverArgs(NamedTuple):
    generic: Callable[["SolverArgs"], NId | None]
    graph: SyntaxGraph
    path: Path
    n_id: NId

    def fork_n_id(self, n_id: NId) -> "SolverArgs":
        return SolverArgs(
            generic=self.generic,
            graph=self.graph,
            path=self.path,
            n_id=n_id,
        )


def solve_symbol_lookup(args: SolverArgs) -> NId | None:
    symbol = args.graph.nodes[args.n_id]["symbol"]
    try:
        search_path = get_lookup_path(args.graph, args.path, args.n_id)
    except ValueError:
        return None

    return definition_search(args.graph, search_path, symbol)


SOLVERS: dict[str, Callable[[SolverArgs], NId | None]] = {
    "SymbolLookup": solve_symbol_lookup,
}


def generic(args: SolverArgs) -> NId | None:
    if args.n_id not in args.graph.nodes:
        return None
    node_type = args.graph.nodes[args.n_id]["label_type"]
    if solver := SOLVERS.get(node_type):
        return solver(args)
    return None


def solve_invocation(graph: SyntaxGraph, path: Path, n_id: NId) -> NId | None:
    return generic(SolverArgs(generic, graph, path, n_id))



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.catch_clause import (
    build_catch_clause_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    block = args.ast_graph.nodes[args.n_id]["label_field_body"]
    param_id = match_ast_d(args.ast_graph, args.n_id, "catch_formal_parameter")
    return build_catch_clause_node(args, block, param_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.metadata.java import (
    add_instance_to_metadata,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    var_type_id = args.ast_graph.nodes[args.n_id]["label_field_type"]
    var_type = node_to_str(args.ast_graph, var_type_id)

    declarator_id = args.ast_graph.nodes[args.n_id]["label_field_declarator"]

    var_id = args.ast_graph.nodes[declarator_id]["label_field_name"]
    var_name = node_to_str(args.ast_graph, var_id)

    value_id = args.ast_graph.nodes[declarator_id].get("label_field_value")

    if args.syntax_graph.nodes.get(NId(0)):
        add_instance_to_metadata(args, var_type, var_name)

    return build_variable_declaration_node(args, var_name, var_type, value_id)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.return_statement import (
    build_return_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    match = match_ast(args.ast_graph, args.n_id, "return", "yield")
    if val_id := match.get("__0__"):
        return build_return_node(args, val_id)
    return build_return_node(args, None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.member_access import (
    build_member_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes
    child_ids = [
        *filter(
            lambda n_id: nodes[n_id].get("label_type") != "line_comment",
            adj_ast(graph, args.n_id),
        ),
    ]
    expected_children_for_member_access = 2
    if (
        len(child_ids) == expected_children_for_member_access
        and graph.nodes[child_ids[1]]["label_type"] == "navigation_suffix"
        and (identifier_id := match_ast(graph, child_ids[1]).get("__1__"))
    ):
        member = node_to_str(graph, identifier_id)
    else:
        member = node_to_str(graph, child_ids[-1])

    expression = node_to_str(graph, child_ids[0])

    return build_member_access_node(args, member, expression, child_ids[0])



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.yaml import (
    block_mapping as yaml_block_mapping,
)
from blends.syntax.readers.yaml import (
    block_mapping_pair as yaml_block_mapping_pair,
)
from blends.syntax.readers.yaml import (
    block_sequence as yaml_block_sequence,
)
from blends.syntax.readers.yaml import (
    flow_node as yaml_flow_node,
)
from blends.syntax.readers.yaml import (
    stream as yaml_stream,
)

YAML_DISPATCHERS: Dispatcher = {
    "block_mapping": yaml_block_mapping.reader,
    "flow_mapping": yaml_block_mapping.reader,
    "block_mapping_pair": yaml_block_mapping_pair.reader,
    "flow_pair": yaml_block_mapping_pair.reader,
    "block_sequence": yaml_block_sequence.reader,
    "flow_node": yaml_flow_node.reader,
    "stream": yaml_stream.reader,
}



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.finally_clause import (
    build_finally_clause_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    childs = match_ast(args.ast_graph, args.n_id, "block")
    finally_block = childs.get("block")
    return build_finally_clause_node(args, finally_block)



from unittest.mock import (
    patch,
)

import pytest

from blends.models import (
    Graph,
    Language,
    NId,
    SyntaxGraph,
)
from blends.stack.attributes import STACK_GRAPH_KIND, STACK_GRAPH_SYMBOL
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.syntax.builders.import_global import (
    build_import_global_node,
)
from blends.syntax.builders.import_module import (
    build_import_module_node,
)
from blends.syntax.builders.import_statement import (
    build_import_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
    SyntaxMetadata,
)


@pytest.fixture  # type: ignore[misc,attr-defined]
def mock_args() -> SyntaxGraphArgs:  # type: ignore[misc]
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {
        "class_path": [],
        "scope_stack": ["1"],
        "next_synthetic_id": 100,
    }
    syntax_graph.add_node("1", label_type="File")

    def generic(args: SyntaxGraphArgs) -> NId:
        return args.n_id

    return SyntaxGraphArgs(
        generic=generic,
        path="test.py",
        language=Language.PYTHON,
        ast_graph=ast_graph,
        syntax_graph=syntax_graph,
        n_id="2",
        metadata=metadata,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_global_binds_alias_or_last_segment(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_import_global_node(
            mock_args,
            expression="pkg.mod",
            module_nodes=set(),
            alias="m",
        )

        assert (
            mock_args.syntax_graph.nodes["2"]["stack_graph_kind"]
            == StackGraphNodeKind.POP_SCOPED_SYMBOL.value
        )
        assert mock_args.syntax_graph.nodes["2"]["stack_graph_symbol"] == "m"
        assert mock_args.syntax_graph.has_edge("1", "2")
        edge_data = mock_args.syntax_graph["1"]["2"]
        assert edge_data["precedence"] == 1


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_module_binds_alias_or_last_segment(
    mock_args: SyntaxGraphArgs,
) -> None:
    module_args = SyntaxGraphArgs(
        generic=mock_args.generic,
        path=mock_args.path,
        language=mock_args.language,
        ast_graph=mock_args.ast_graph,
        syntax_graph=mock_args.syntax_graph,
        n_id="3",
        metadata=mock_args.metadata,
    )
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_import_module_node(
            module_args,
            expression="pkg.mod",
            alias=None,
        )

        assert (
            mock_args.syntax_graph.nodes["3"]["stack_graph_kind"]
            == StackGraphNodeKind.POP_SCOPED_SYMBOL.value
        )
        assert mock_args.syntax_graph.nodes["3"]["stack_graph_symbol"] == "mod"
        assert mock_args.syntax_graph.has_edge("1", "3")


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_statement_multiple_marks_each_child_as_definition(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_import_statement_node(
            mock_args,
            {"corrected_n_id": "10", "expression": "a.b", "label_alias": "c"},
            {"corrected_n_id": "11", "expression": "x.y", "label_alias": ""},
        )

        assert mock_args.syntax_graph.has_edge("2", "10")
        assert mock_args.syntax_graph.has_edge("2", "11")

        assert (
            mock_args.syntax_graph.nodes["10"]["stack_graph_kind"]
            == StackGraphNodeKind.POP_SCOPED_SYMBOL.value
        )
        assert mock_args.syntax_graph.nodes["10"]["stack_graph_symbol"] == "c"
        assert mock_args.syntax_graph.has_edge("1", "10")

        assert (
            mock_args.syntax_graph.nodes["11"]["stack_graph_kind"]
            == StackGraphNodeKind.POP_SCOPED_SYMBOL.value
        )
        assert mock_args.syntax_graph.nodes["11"]["stack_graph_symbol"] == "y"
        assert mock_args.syntax_graph.has_edge("1", "11")


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_nodes_do_not_emit_stack_graph_when_feature_preview_disabled(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", return_value=False):
        build_import_global_node(
            mock_args,
            expression="pkg.mod",
            module_nodes=set(),
            alias="m",
        )

        assert "stack_graph_kind" not in mock_args.syntax_graph.nodes["2"]
        assert "stack_graph_symbol" not in mock_args.syntax_graph.nodes["2"]
        assert not mock_args.syntax_graph.has_edge("1", "2")


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_global_creates_reference_chain_to_root(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_import_global_node(
            mock_args,
            expression="pkg.mod",
            module_nodes=set(),
            alias=None,
        )

    root_nid = mock_args.metadata.get("root_nid")
    assert root_nid is not None
    assert root_nid in mock_args.syntax_graph.nodes
    graph = mock_args.syntax_graph
    import_successors = [
        s
        for s in graph.successors("2")
        if graph.nodes[s].get(STACK_GRAPH_KIND) == StackGraphNodeKind.PUSH_SYMBOL.value
    ]
    assert len(import_successors) == 1
    push_node = import_successors[0]
    assert graph.nodes[push_node][STACK_GRAPH_SYMBOL] == "mod"
    assert graph.has_edge(push_node, root_nid)


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_module_creates_reference_chain_to_root(
    mock_args: SyntaxGraphArgs,
) -> None:
    module_args = SyntaxGraphArgs(
        generic=mock_args.generic,
        path=mock_args.path,
        language=mock_args.language,
        ast_graph=mock_args.ast_graph,
        syntax_graph=mock_args.syntax_graph,
        n_id="3",
        metadata=mock_args.metadata,
    )
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_import_module_node(
            module_args,
            expression="pkg.mod",
            alias=None,
        )

    root_nid = mock_args.metadata.get("root_nid")
    assert root_nid is not None
    graph = mock_args.syntax_graph
    import_successors = [
        s
        for s in graph.successors("3")
        if graph.nodes[s].get(STACK_GRAPH_KIND) == StackGraphNodeKind.PUSH_SYMBOL.value
    ]
    assert len(import_successors) == 1
    push_node = import_successors[0]
    assert graph.nodes[push_node][STACK_GRAPH_SYMBOL] == "mod"
    assert graph.has_edge(push_node, root_nid)


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_statement_creates_reference_chain_to_root(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_import_statement_node(
            mock_args,
            {"corrected_n_id": "10", "expression": "a.b"},
            {"corrected_n_id": "11", "expression": "x.y"},
        )

    root_nid = mock_args.metadata.get("root_nid")
    assert root_nid is not None
    graph = mock_args.syntax_graph
    for node_id, expected_symbol in [("10", "b"), ("11", "y")]:
        push_nodes = [
            s
            for s in graph.successors(node_id)
            if graph.nodes[s].get(STACK_GRAPH_KIND) == StackGraphNodeKind.PUSH_SYMBOL.value
        ]
        assert len(push_nodes) == 1
        assert graph.nodes[push_nodes[0]][STACK_GRAPH_SYMBOL] == expected_symbol
        assert graph.has_edge(push_nodes[0], root_nid)


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_alias_does_not_affect_reference_chain(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_import_global_node(
            mock_args,
            expression="pkg.mod",
            module_nodes=set(),
            alias="m",
        )

    graph = mock_args.syntax_graph
    assert graph.nodes["2"][STACK_GRAPH_SYMBOL] == "m"

    push_nodes = [
        s
        for s in graph.successors("2")
        if graph.nodes[s].get(STACK_GRAPH_KIND) == StackGraphNodeKind.PUSH_SYMBOL.value
    ]
    assert len(push_nodes) == 1
    assert graph.nodes[push_nodes[0]][STACK_GRAPH_SYMBOL] == "mod"


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_global_from_import_chain_uses_module_name(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_import_global_node(
            mock_args,
            expression="mf_core.SENTINEL",
            module_nodes=set(),
            alias=None,
            module_name="mf_core",
        )

    graph = mock_args.syntax_graph
    assert graph.nodes["2"][STACK_GRAPH_SYMBOL] == "SENTINEL"
    push_nodes = [
        s
        for s in graph.successors("2")
        if graph.nodes[s].get(STACK_GRAPH_KIND) == StackGraphNodeKind.PUSH_SYMBOL.value
    ]
    assert len(push_nodes) == 1
    assert graph.nodes[push_nodes[0]][STACK_GRAPH_SYMBOL] == "mf_core"


@pytest.mark.blends_test_group("stack_unittesting")
def test_import_global_from_import_with_alias_chain_uses_module_name(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_import_global_node(
            mock_args,
            expression="mf_core.CHUNK_SIZE",
            module_nodes=set(),
            alias="BUFFER_SIZE",
            module_name="mf_core",
        )

    graph = mock_args.syntax_graph
    assert graph.nodes["2"][STACK_GRAPH_SYMBOL] == "BUFFER_SIZE"
    push_nodes = [
        s
        for s in graph.successors("2")
        if graph.nodes[s].get(STACK_GRAPH_KIND) == StackGraphNodeKind.PUSH_SYMBOL.value
    ]
    assert len(push_nodes) == 1
    assert graph.nodes[push_nodes[0]][STACK_GRAPH_SYMBOL] == "mf_core"



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.argument_list import (
    build_argument_list_node,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)

    if (
        len(c_ids) == 1
        and (literal := match_ast_d(graph, args.n_id, "interpreted_string_literal"))
        and (value := graph.nodes[literal].get("label_text"))
    ):
        return build_string_literal_node(args, value)

    return build_argument_list_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"] not in {",", ":"}),
    )



from collections.abc import (
    Iterable,
)

from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_try_statement_node(
    args: SyntaxGraphArgs,
    block_node: NId,
    catch_blocks: Iterable[NId] | None,
    try_block: NId | None,
    resources_spec: NId | None,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        block_id=block_node,
        label_type="TryStatement",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(block_node)),
        label_ast="AST",
    )

    if resources_spec:
        args.syntax_graph.update_node_attribute(args.n_id, "resources_id", resources_spec)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(resources_spec)),
            label_ast="AST",
        )

    if catch_blocks:
        for c_id in catch_blocks:
            args.syntax_graph.add_edge(
                args.n_id,
                args.generic(args.fork_n_id(c_id)),
                label_ast="AST",
            )

    if try_block:
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(try_block)),
            label_ast="AST",
        )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs = match_ast(graph, args.n_id, "=")
    var_name = node_to_str(graph, var_id) if (var_id := childs.get("__0__")) else "UnnamedVar"

    value_id = childs.get("__1__")

    if value_id and graph.nodes[value_id]["label_type"] == "jsx_expression":
        value_id = match_ast(graph, value_id, "{").get("__0__")

    return build_variable_declaration_node(args, var_name, None, value_id)



from __future__ import annotations

from typing import TYPE_CHECKING

from blends.stack.policies.java import JavaExportPolicy

if TYPE_CHECKING:
    from blends.models import NId, SyntaxGraph


class KotlinExportPolicy(JavaExportPolicy):
    def get_visibility_modifier(
        self,
        graph: SyntaxGraph,
        node_id: NId,
    ) -> str | None:
        modifier = super().get_visibility_modifier(graph, node_id)
        if modifier is None:
            return "public"
        return modifier



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_group_d,
)
from blends.syntax.builders.expression_statement import (
    build_expression_statement_node,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    var_ids = match_ast_group_d(graph, args.n_id, "variable_declarator")
    if len(var_ids) == 1:
        declared_var = var_ids[0]
        var_id = graph.nodes[declared_var]["label_field_name"]
        var_name = node_to_str(graph, var_id)
        var_name = var_name[1:-1] if var_name.startswith("{") else var_name
        value_id = graph.nodes[declared_var].get("label_field_value")
        return build_variable_declaration_node(args, var_name, None, value_id)

    return build_expression_statement_node(args, iter(var_ids))



from blends.models import (
    NId,
)
from blends.query import (
    get_node_by_path,
)
from blends.syntax.builders.for_statement import (
    build_for_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    body_id = n_attrs["label_field_body"]

    var_node = get_node_by_path(graph, args.n_id, "enumerators", "enumerator", "identifier")

    condition_node = get_node_by_path(
        graph,
        args.n_id,
        "enumerators",
        "enumerator",
        "infix_expression",
    )

    return build_for_statement_node(args, var_node, condition_node, None, body_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.element_access import (
    build_element_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    n_attrs = args.ast_graph.nodes[args.n_id]
    operand_id = n_attrs["label_field_operand"]
    start_id = n_attrs.get("label_field_start")

    return build_element_access_node(args, operand_id, start_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.method_invocation import (
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expr_id = graph.nodes[args.n_id]["label_field_field"]
    expr = node_to_str(graph, expr_id)
    return build_method_invocation_node(
        args,
        expr,
        {
            "expression_id": expr_id,
            "object_id": graph.nodes[args.n_id]["label_field_value"],
        },
    )



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.import_global import (
    build_import_global_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import node_to_str


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    alias: str | None = None

    exp_n_id = match_ast_d(args.ast_graph, args.n_id, "interpreted_string_literal")
    if exp_n_id:
        expression = graph.nodes[exp_n_id].get("label_text", "")[1:-1]
    else:
        expression = node_to_str(graph, args.n_id)

    if alias_n_id := match_ast_d(args.ast_graph, args.n_id, "package_identifier"):
        alias = graph.nodes[alias_n_id].get("label_text")

    return build_import_global_node(args, expression, set(), alias)



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        label_type="Continue",
    )
    return args.n_id



import json
from pathlib import Path

import pytest

from blends import get_graphs_from_path
from blends.models import Graph


def export_graph_as_json(graph: Graph) -> dict[str, dict[str, dict[str, dict[str, str]]]]:
    data: dict[str, dict[str, dict[str, dict[str, str]]]] = {}
    data["nodes"] = {}
    data["edges"] = {}

    for n_id, n_attrs in graph.nodes.items():
        data["nodes"][n_id] = n_attrs.copy()

    for n_id_from, n_id_to in graph.edges:
        data["edges"].setdefault(n_id_from, {})
        data["edges"][n_id_from][n_id_to] = graph[n_id_from][n_id_to].copy()

    return data


@pytest.mark.blends_test_group("functional")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    ("test_file", "expected_suffix"),
    [
        ("c_sharp.cs", "c_sharp"),
        ("dart.dart", "dart"),
        ("go.go", "go"),
        ("terraform.tf", "hcl"),
        ("java.java", "java"),
        ("javascript.js", "javascript"),
        ("json.json", "json"),
        ("kotlin.kt", "kotlin"),
        ("php.php", "php"),
        ("python.py", "python"),
        ("ruby.rb", "ruby"),
        ("scala.scala", "scala"),
        ("swift.swift", "swift"),
        ("syntax_cfg.ts", "typescript"),
        ("yaml.yaml", "yaml"),
    ],
)
def test_syntax_graph_generation(test_file: str, expected_suffix: str) -> None:  # type: ignore[misc]
    test_data_dir = Path(__file__).parent.parent / "data"
    test_file_path = test_data_dir / "test_files/syntax_graph" / test_file

    graphs = get_graphs_from_path(test_file_path)

    assert graphs.ast_graph is not None, f"AST graph is None for {test_file}"
    assert graphs.syntax_graph is not None, f"Syntax graph is None for {test_file}"

    relative_test_path = f"test/data/test_files/{test_file}"

    graph_db: dict[str, dict[str, dict[str, Graph]]] = {
        "graphs": {
            relative_test_path: {
                "graph": graphs.ast_graph,
                "syntax_graph": graphs.syntax_graph,
            },
        },
    }

    graph_db_as_json_str = json.dumps(
        graph_db,
        indent=2,
        sort_keys=True,
        default=export_graph_as_json,
    )

    expected_path = test_data_dir / "results" / f"root-graph_{expected_suffix}.json"

    if expected_path.exists():
        with expected_path.open(encoding="utf-8") as handle:
            expected_graph = handle.read()

        if expected_graph != graph_db_as_json_str:
            with expected_path.open("w", encoding="utf-8") as handle_w:
                handle_w.write(graph_db_as_json_str)
            pytest.fail(f"Graph DB mismatch for suffix: {expected_suffix}")  # type: ignore[attr-defined]
    else:
        expected_path.parent.mkdir(parents=True, exist_ok=True)
        with expected_path.open("w", encoding="utf-8") as handle:
            handle.write(graph_db_as_json_str)
        pytest.fail(  # type: ignore[attr-defined]
            f"Expected file did not exist, created it: {expected_path}. "
            "Please review and commit if correct."
        )



from blends.models import (
    NId,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    condition = n_attrs["label_field_value"]
    if_nil = n_attrs.get("label_field_if_nil")
    return build_if_node(args, condition, if_nil, None)



from dataclasses import dataclass
from typing import TYPE_CHECKING

from blends.stack.partial_path.bindings import (
    PartialScopeStackBindings,
    PartialSymbolStackBindings,
)
from blends.stack.partial_path.errors import (
    _raise_incorrect_source_node,
    _raise_unbound_scope_stack_variable,
    _raise_unbound_symbol_stack_variable,
)
from blends.stack.partial_path.half_open import (
    halfopen_closed_partial_postcondition,
    halfopen_closed_partial_precondition,
)

if TYPE_CHECKING:
    from blends.stack.partial_path.partial_stacks import PartialScopeStack, PartialSymbolStack
    from blends.stack.partial_path.variables import ScopeStackVariable
    from blends.stack.selection import PartialPathEdge
    from blends.stack.view import StackGraphView


def _scope_vars_from_scope_stack(scope_stack: "PartialScopeStack") -> set["ScopeStackVariable"]:
    if scope_stack.variable is None:
        return set()
    return {scope_stack.variable}


def _scope_vars_from_symbol_stack(symbol_stack: "PartialSymbolStack") -> set["ScopeStackVariable"]:
    vars_from_symbols: set[ScopeStackVariable] = set()
    for symbol in symbol_stack.symbols:
        scopes = symbol.scopes
        if scopes is None or scopes.variable is None:
            continue
        vars_from_symbols.add(scopes.variable)
    return vars_from_symbols


@dataclass(frozen=True, slots=True)
class PartialPath:
    start_node_index: int
    end_node_index: int
    symbol_stack_precondition: "PartialSymbolStack"
    symbol_stack_postcondition: "PartialSymbolStack"
    scope_stack_precondition: "PartialScopeStack"
    scope_stack_postcondition: "PartialScopeStack"
    edges: tuple["PartialPathEdge", ...] = ()

    def largest_symbol_stack_var(self) -> int:
        return max(
            self.symbol_stack_precondition.largest_symbol_stack_var(),
            self.symbol_stack_postcondition.largest_symbol_stack_var(),
        )

    def largest_scope_stack_var(self) -> int:
        return max(
            self.scope_stack_precondition.largest_scope_stack_var(),
            self.scope_stack_postcondition.largest_scope_stack_var(),
            self.symbol_stack_precondition.largest_scope_stack_var(),
            self.symbol_stack_postcondition.largest_scope_stack_var(),
        )

    def ensure_no_overlapping_variables(self, other: "PartialPath") -> "PartialPath":
        symbol_offset = other.largest_symbol_stack_var()
        scope_offset = other.largest_scope_stack_var()
        if symbol_offset == 0 and scope_offset == 0:
            return self
        return PartialPath(
            start_node_index=self.start_node_index,
            end_node_index=self.end_node_index,
            symbol_stack_precondition=self.symbol_stack_precondition.with_offset(
                symbol_offset, scope_offset
            ),
            symbol_stack_postcondition=self.symbol_stack_postcondition.with_offset(
                symbol_offset, scope_offset
            ),
            scope_stack_precondition=self.scope_stack_precondition.with_offset(scope_offset),
            scope_stack_postcondition=self.scope_stack_postcondition.with_offset(scope_offset),
            edges=self.edges,
        )

    def validate_bound_vars(self) -> None:
        pre_symbol_var = self.symbol_stack_precondition.variable
        post_symbol_var = self.symbol_stack_postcondition.variable
        if post_symbol_var is not None and post_symbol_var != pre_symbol_var:
            _raise_unbound_symbol_stack_variable()

        pre_scope_vars = _scope_vars_from_scope_stack(self.scope_stack_precondition)
        pre_scope_vars.update(_scope_vars_from_symbol_stack(self.symbol_stack_precondition))
        post_scope_vars = _scope_vars_from_scope_stack(self.scope_stack_postcondition)
        post_scope_vars.update(_scope_vars_from_symbol_stack(self.symbol_stack_postcondition))
        if not post_scope_vars.issubset(pre_scope_vars):
            _raise_unbound_scope_stack_variable()

    def concatenate(self, view: "StackGraphView", rhs: "PartialPath") -> "PartialPath":
        if self.end_node_index != rhs.start_node_index:
            _raise_incorrect_source_node()
        rhs = rhs.ensure_no_overlapping_variables(self)
        lhs_post = halfopen_closed_partial_postcondition(
            view,
            self.end_node_index,
            self.symbol_stack_postcondition,
            self.scope_stack_postcondition,
        )
        rhs_pre = halfopen_closed_partial_precondition(
            view,
            rhs.start_node_index,
            rhs.symbol_stack_precondition,
            rhs.scope_stack_precondition,
        )
        symbol_bindings = PartialSymbolStackBindings.new()
        scope_bindings = PartialScopeStackBindings.new()
        lhs_post.symbol_stack.unify(rhs_pre.symbol_stack, symbol_bindings, scope_bindings)
        lhs_post.scope_stack.unify(rhs_pre.scope_stack, scope_bindings)
        result = PartialPath(
            start_node_index=self.start_node_index,
            end_node_index=rhs.end_node_index,
            symbol_stack_precondition=self.symbol_stack_precondition.apply_bindings(
                symbol_bindings, scope_bindings
            ),
            symbol_stack_postcondition=rhs.symbol_stack_postcondition.apply_bindings(
                symbol_bindings, scope_bindings
            ),
            scope_stack_precondition=self.scope_stack_precondition.apply_bindings(scope_bindings),
            scope_stack_postcondition=rhs.scope_stack_postcondition.apply_bindings(scope_bindings),
            edges=self.edges + rhs.edges,
        )
        result.validate_bound_vars()
        return result



from typing import TypedDict, cast

from blends.ctx import ctx
from blends.models import (
    Language,
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    pop_scoped_symbol_node_attributes,
    scope_node_attributes,
)
from blends.stack.scope_types import ScopeType
from blends.syntax.builders.utils import (
    get_next_binding_precedence,
    get_next_synthetic_node_id,
)
from blends.syntax.metadata.java import (
    add_node_range_to_method,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


class DirectChildren(TypedDict, total=False):
    block_id: NId
    modifiers_id: NId
    parameters_id: NId


class ListChildren(TypedDict, total=False):
    modifiers_ids: list[NId]
    parameters_ids: list[NId]


def _process_method_children(
    args: SyntaxGraphArgs,
    children: DirectChildren,
) -> None:
    for name_node, child_id in cast("dict[str, NId]", children).items():
        args.syntax_graph.update_node_attribute(args.n_id, name_node, child_id)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(child_id)),
            label_ast="AST",
        )


def _process_method_list_children(
    args: SyntaxGraphArgs,
    children: ListChildren,
) -> None:
    for n_ids in cast("dict[str, list[NId]]", children).values():
        for child_id in n_ids:
            args.syntax_graph.add_edge(
                args.n_id,
                args.generic(args.fork_n_id(child_id)),
                label_ast="AST",
            )


def _setup_stack_graph_scope(args: SyntaxGraphArgs, name: str | None) -> None:
    scope_stack = args.metadata.setdefault("scope_stack", [])
    if name and scope_stack and (parent_scope := scope_stack[-1]):
        precedence = get_next_binding_precedence(args)
        method_def_id = get_next_synthetic_node_id(args)
        args.syntax_graph.add_node(
            method_def_id,
            label_type="MethodDeclaration",
            name=name,
            sg_node_type="definition",
        )
        args.syntax_graph.update_node(
            method_def_id,
            pop_scoped_symbol_node_attributes(
                symbol=name,
                precedence=precedence,
            ),
        )
        add_edge(
            args.syntax_graph,
            Edge(source=parent_scope, sink=method_def_id, precedence=precedence),
        )
        add_edge(
            args.syntax_graph,
            Edge(source=method_def_id, sink=args.n_id, precedence=0),
        )
        ast_node = args.ast_graph.nodes.get(args.n_id, {})
        if (label_l := ast_node.get("label_l")) is not None:
            args.syntax_graph.update_node_attribute(method_def_id, "label_l", label_l)
            args.syntax_graph.update_node_attribute(
                method_def_id, "label_c", ast_node.get("label_c", "0")
            )

    args.syntax_graph.update_node(
        args.n_id,
        scope_node_attributes(is_exported=False, scope_type=ScopeType.FUNCTION),
    )
    if scope_stack:
        add_edge(
            args.syntax_graph,
            Edge(source=args.n_id, sink=scope_stack[-1], precedence=0),
        )
    scope_stack.append(args.n_id)


def _cleanup_stack_graph_scope(args: SyntaxGraphArgs) -> None:
    scope_stack = args.metadata.setdefault("scope_stack", [])
    if scope_stack and scope_stack[-1] == args.n_id:
        scope_stack.pop()


def build_method_declaration_node(
    args: SyntaxGraphArgs,
    name: str | None,
    direct_children: DirectChildren,
    list_children: ListChildren,
    access_modifiers: str | None = None,
) -> NId:
    args.syntax_graph.add_node(args.n_id, label_type="MethodDeclaration")
    if name:
        args.syntax_graph.update_node_attribute(args.n_id, "name", name)

    if access_modifiers:
        args.syntax_graph.update_node_attribute(args.n_id, "access_modifiers", access_modifiers)

    if ctx.has_feature_flag("StackGraph"):
        _setup_stack_graph_scope(args, name)

    _process_method_children(args, direct_children)
    _process_method_list_children(args, list_children)

    if args.language == Language.JAVA and args.syntax_graph.nodes.get(NId(0)) and name:
        add_node_range_to_method(args, name)

    if ctx.has_feature_flag("StackGraph"):
        _cleanup_stack_graph_scope(args)

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes
    n_attrs = nodes[args.n_id]

    ignored_types = {"(", ",", ")"}

    expr_id = n_attrs["label_field_scope"]
    if nodes[expr_id]["label_type"] == "relative_scope" and (c_ids := adj_ast(graph, expr_id)):
        expr_id = c_ids[0]
    raw_expr = node_to_str(graph, expr_id)
    method_name_n_id = n_attrs["label_field_name"]
    method_name = nodes[method_name_n_id].get("label_text")
    expr = f"{raw_expr}::{method_name}"

    arguments_id = graph.nodes[args.n_id].get("label_field_arguments")
    if arguments_id:
        arguments = {
            _id
            for _id in adj_ast(graph, arguments_id)
            if graph.nodes[_id].get("label_type") not in ignored_types
        }
        if not arguments:
            arguments_id = None

    children: DirectChildren = {"expression_id": expr_id}
    if arguments_id:
        children["arguments_id"] = arguments_id

    return build_method_invocation_node(args, expr, children)



from blends.models import (
    Graph,
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_d,
    match_ast_group_d,
    pred_ast,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def get_false_id(graph: Graph, if_n_id: NId, current_n_id: NId) -> NId | None:
    for elif_n_id in match_ast_group_d(graph, if_n_id, "else_if_clause"):
        if int(elif_n_id) > int(current_n_id):
            return elif_n_id
    if (else_n_id := match_ast_d(graph, if_n_id, "else_clause")) and (
        false_id := graph.nodes[else_n_id].get("label_field_body")
    ):
        return false_id if isinstance(false_id, str) else None
    return None


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes
    condition_id = graph.nodes[args.n_id]["label_field_condition"]

    if nodes[condition_id].get("label_type") == "parenthesized_expression":
        c_ids = match_ast(graph, condition_id)
        condition_id = c_ids.get("__1__")

    true_id = graph.nodes[args.n_id]["label_field_body"]

    if not (false_id := graph.nodes[args.n_id].get("label_field_alternative")):
        p_n_ids = pred_ast(graph, args.n_id, -1)
        for n_id in p_n_ids:
            if (nodes[n_id].get("label_type") == "if_statement") and (
                args.n_id in adj_ast(graph, n_id)
            ):
                false_id = get_false_id(graph, n_id, args.n_id)
    if false_id and nodes[false_id].get("label_type") == "else_clause":
        false_id = graph.nodes[false_id].get("label_field_body")

    return build_if_node(args, condition_id, true_id, false_id)



from pathlib import PurePosixPath


def compute_module_path_segments(file_path: str) -> list[str]:
    posix = PurePosixPath(file_path)
    stem = posix.stem

    if stem == "__init__":
        return []

    return [stem]



from collections.abc import Callable
from typing import NamedTuple, NotRequired, TypedDict

from blends.models import (
    Graph,
    Language,
    NId,
    SyntaxGraph,
)


class SyntaxMetadata(TypedDict):
    class_path: list[str]
    scope_stack: list[NId]
    class_member_scope_stack: NotRequired[list[tuple[NId, str]]]
    binding_counter: NotRequired[dict[NId, int]]
    next_synthetic_id: NotRequired[int]
    root_nid: NotRequired[NId]
    is_classmethod: NotRequired[bool]


class SyntaxGraphArgs(NamedTuple):
    generic: Callable[["SyntaxGraphArgs"], NId]
    path: str
    language: Language
    ast_graph: Graph
    syntax_graph: SyntaxGraph
    n_id: NId
    metadata: SyntaxMetadata

    def fork_n_id(self, n_id: NId) -> "SyntaxGraphArgs":
        return SyntaxGraphArgs(
            generic=self.generic,
            path=self.path,
            language=self.language,
            ast_graph=self.ast_graph,
            syntax_graph=self.syntax_graph,
            n_id=n_id,
            metadata=self.metadata,
        )


SyntaxReader = Callable[[SyntaxGraphArgs], NId]
Dispatcher = dict[str, SyntaxReader]


class FileStructData(TypedDict):
    node: NId
    type: str
    data: dict[str, str] | str


class FileInstanceData(TypedDict):
    object: str
    source: str
    source_type: str


class ReaderLogicError(Exception):
    pass


class SyntaxCfgArgs(NamedTuple):
    generic: Callable[["SyntaxCfgArgs"], NId]
    graph: SyntaxGraph
    n_id: NId
    nxt_id: NId | None
    is_multifile: bool = False

    def fork(self, n_id: NId, nxt_id: NId | None) -> "SyntaxCfgArgs":
        return SyntaxCfgArgs(
            generic=self.generic,
            graph=self.graph,
            n_id=n_id,
            nxt_id=nxt_id,
            is_multifile=self.is_multifile,
        )


class MissingCfgBuilderError(Exception):
    pass



from blends.models import (
    NId,
)
from blends.query import (
    get_node_by_path,
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    class_node = args.ast_graph.nodes[args.n_id]
    if declaration_n_id := match_ast_d(args.ast_graph, args.n_id, "identifier"):
        declaration_node = args.ast_graph.nodes[declaration_n_id]
        declaration_line = declaration_node["label_l"]
        args.ast_graph.nodes[args.n_id]["label_l"] = declaration_line
    name_id = class_node["label_field_name"]
    block_id = class_node["label_field_body"]
    name = node_to_str(args.ast_graph, name_id)
    attrl_ids = match_ast_group_d(args.ast_graph, args.n_id, "attribute_list")
    inherited_id = get_node_by_path(args.ast_graph, args.n_id, "base_list", "identifier")
    inherited_class = node_to_str(args.ast_graph, inherited_id) if inherited_id else None

    return build_class_node(args, name, block_id, attrl_ids, inherited_class)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.while_statement import (
    build_while_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)

PARENTHESIZED_EXPRESSION_CHILDREN_COUNT = 3


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    match_childs = match_ast(
        args.ast_graph,
        args.n_id,
        "parenthesized_expression",
        "statement_block",
    )
    block_id = match_childs.get("statement_block")
    if not block_id:
        block_id = adj_ast(graph, args.n_id)[-1]

    if block_id and graph.nodes[block_id]["label_type"] == "expression_statement":
        block_id = adj_ast(graph, block_id)[0]

    conditional_id = match_childs.get("parenthesized_expression")
    if conditional_id:
        adj_nodes = adj_ast(graph, conditional_id)
        if len(adj_nodes) == PARENTHESIZED_EXPRESSION_CHILDREN_COUNT:
            conditional_id = adj_nodes[1]
        else:
            conditional_id = adj_nodes[0] if adj_nodes else None

    return build_while_statement_node(args, block_id, conditional_id)



from blends.models import (
    NId,
)
from blends.query import (
    pred_ast,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    parent_id = pred_ast(args.ast_graph, args.n_id)[0]
    symbol = node_to_str(args.ast_graph, args.n_id)
    parent_node = args.ast_graph.nodes[parent_id]
    if parent_node.get("label_type") == "formal_parameters":
        return build_parameter_node(args=args, variable=symbol, variable_type=None, value_id=None)
    return build_symbol_lookup_node(args, symbol)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.object_creation import (
    build_object_creation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    node_attr = graph.nodes[args.n_id]
    type_id = node_attr["label_field_type"]
    name = node_to_str(graph, type_id)

    if arguments_id := node_attr.get("label_field_arguments"):
        if "__0__" not in match_ast(graph, arguments_id, "(", ")"):
            arguments_id = None
        return build_object_creation_node(args, name, arguments_id, None)

    init_id = node_attr.get("label_field_initializer")
    if (
        init_id
        and (childs := match_ast(graph, init_id, "assignment_expression"))
        and not childs.get("__0__")
    ):
        init_id = None

    return build_object_creation_node(args, name, None, init_id)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from blends.models import NId, SyntaxGraph


class ExportPolicy(ABC):
    @abstractmethod
    def is_scope_exported(
        self,
        graph: SyntaxGraph,
        node_id: NId,
        scope_type: str | None,
        /,
    ) -> bool: ...

    @abstractmethod
    def is_definition_exported(
        self,
        graph: SyntaxGraph,
        node_id: NId,
        parent_scope_id: NId | None,
        /,
    ) -> bool: ...

    def get_visibility_modifier(
        self,
        graph: SyntaxGraph,
        node_id: NId,
    ) -> str | None:
        modifiers = graph.nodes[node_id].get(
            "access_modifiers",
            graph.nodes[node_id].get("access_modifier", ""),
        )
        for modifier in ("public", "private", "protected", "internal"):
            if modifier in modifiers:
                return modifier
        return None



import logging
import re

from blends.models import Content, CustomParser
from blends.tree.parse import ParsingError, get_tree

LOGGER = logging.getLogger(__name__)

HELM_EXPR_PATTERN = re.compile(r"\{\{\-?\s*[^{}]*?\s*\-?\}\}")
KV_PATTERN = re.compile(r"^(?P<indent>\s*)(?P<dash>- )?(?P<key>[^:]+:\s*)(?P<value>.*)$")
BLOCK_HELM_COMMENT_PATTERN = re.compile(r"\{\{\s*-?\s*/\*([\s\S]*?)\*/\s*-?\s*\}\}")


def _validate_helm_required_keys(content: str) -> bool:
    config_required_keys = ["apiVersion", "kind", "metadata"]
    return all(key in content for key in config_required_keys)


def _convert_block_helm_comment(content: str) -> str:
    def _replacer(match: re.Match[str]) -> str:
        block = match.group(0)
        lines = block.splitlines()
        return "\n".join("# " + line for line in lines)

    return BLOCK_HELM_COMMENT_PATTERN.sub(_replacer, content)


def _normalize_helm_expression(value: str) -> str:
    return re.sub(r"\{\{\-?\s*(.*?)\s*\-?\}\}", r"{{ \1 }}", value)


def _is_full_helm_expression(value: str) -> bool:
    return bool(re.fullmatch(r"\{\{\s*[^{}]+\s*\}\}", value.strip()))


def _has_helm_expression_in_key(key: str) -> bool:
    return bool(HELM_EXPR_PATTERN.search(key))


def _fix_inner_quotes_in_helm(value: str) -> str:
    def _replacer(match: re.Match[str]) -> str:
        expr = match.group(0)
        inner = expr[2:-2]
        fixed_inner = re.sub(r'(?<!\\)"([^"]*?)"', r"'\1'", inner)
        return "{{" + fixed_inner + "}}"

    return HELM_EXPR_PATTERN.sub(_replacer, value)


def _process_key_value_line(line: str) -> str:
    match = KV_PATTERN.match(line)
    if not match:
        return line

    indent = match.group("indent")
    dash = match.group("dash") or ""
    key_part = match.group("key")
    value = match.group("value").strip()

    key_name = key_part.rstrip(": ").strip()

    if _has_helm_expression_in_key(key_name):
        key_name = _normalize_helm_expression(key_name)
        key_part = f"'{key_name}': "
    else:
        key_part = f"{key_name}: "

    if value == "":
        return f"{indent}{dash}{key_part}"

    if HELM_EXPR_PATTERN.search(value):
        value = _fix_inner_quotes_in_helm(value)
        value = _normalize_helm_expression(value)
        min_helm_expressions_in_value = 2
        if (
            ":" in value
            and value.count("{{") >= min_helm_expressions_in_value
            and not (value.startswith('"') and value.endswith('"'))
            and not (value.startswith("'") and value.endswith("'"))
        ):
            value = f'"{value}"'

        if value.startswith('"') and value.endswith('"'):
            return f"{indent}{dash}{key_part}{value}"
        if '"' in value:
            return f"{indent}{dash}{key_part}'{value}'"
        return f'{indent}{dash}{key_part}"{value}"'

    return f"{indent}{dash}{key_part}{value}"


def _process_full_helm_line(line: str) -> str:
    expr = _normalize_helm_expression(line.strip())
    if _is_full_helm_expression(expr):
        indent = len(line) - len(line.lstrip())
        return f"{' ' * indent}# {expr}"
    return line


def _preprocess_helm_templates(content: Content) -> Content | None:
    try:
        content_str = _convert_block_helm_comment(content.content_str)
        lines = content_str.splitlines()
        processed_lines = []

        for line in lines:
            if re.match(r"^\s*-?\s*[\w\-_.\/{}$]+\s*:", line):
                processed_lines.append(_process_key_value_line(line))

            elif _is_full_helm_expression(line.strip()):
                processed_lines.append(_process_full_helm_line(line))

            elif HELM_EXPR_PATTERN.search(line):
                processed_lines.append(
                    f"{_get_indent(line)}# {_normalize_helm_expression(line.strip())}"
                )
            else:
                processed_lines.append(line)
        fixed_content = "\n".join(processed_lines).strip()

        content_bytes = fixed_content.encode("utf-8")
    except Exception as exc:
        LOGGER.exception(
            "Content fix logic got %s error when parsing %s", type(exc).__name__, content.path
        )
        return None

    return Content(content_bytes, fixed_content, content.language, content.path)


def _get_indent(line: str) -> str:
    return line[: len(line) - len(line.lstrip())]


def validator(content: Content) -> bool:
    valid_extensions = {".json", ".yml", ".yaml"}
    return (
        "templates" in str(content.path)
        and any(str(content.path).lower().endswith(ext) for ext in valid_extensions)
        and _validate_helm_required_keys(content.content_str)
    )


def parser(
    content: Content,
) -> Content | None:
    if not HELM_EXPR_PATTERN.search(content.content_str):
        return content

    try:
        get_tree(content)
    except ParsingError:
        fixed_content = _preprocess_helm_templates(content)
    else:
        return content

    if fixed_content is not None:
        try:
            get_tree(fixed_content)
        except ParsingError:
            LOGGER.exception(
                "HELM template still not parseable by tree-sitter after custom fix: %s",
                content.path,
            )
        else:
            return fixed_content

    return None


helm_parser = CustomParser(validator=validator, parser=parser)



from collections.abc import (
    Iterator,
)

from blends.query import (
    adj_cfg,
)
from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def search(args: SearchArgs) -> Iterator[SearchResult]:
    for c_id in adj_cfg(args.graph, args.n_id):
        if (
            args.graph.nodes[c_id]["label_type"] == "VariableDeclaration"
            and args.graph.nodes[c_id]["variable"] == args.symbol
        ):
            yield True, c_id
            break



from blends.models import (
    Graph,
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
    pred_ast,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.builders.selector import (
    build_selector_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def _reader_invocation_expression(
    graph: Graph,
    args: SyntaxGraphArgs,
    arg_part: NId | None,
    id_name: NId | None,
    pred_id: NId | None,
) -> NId | None:
    arg_list = None
    if arg_part:
        arg_list = match_ast_d(graph, arg_part, "arguments")
    if id_name:
        child_id = match_ast_d(graph, id_name, "identifier")
        if pred_id and child_id:
            selector_name = ".".join(
                graph.nodes[_id]["label_text"]
                for _id in [pred_id, child_id]
                if "label_text" in graph.nodes[_id]
            )
            children: DirectChildren = {}
            if child_id:
                children["expression_id"] = child_id
            if arg_list:
                children["arguments_id"] = arg_list
            return build_method_invocation_node(args, selector_name, children)
    return None


def _get_next_arg_part_sibling(
    graph: Graph,
    n_id: NId,
    parent_id: NId,
) -> NId | None:
    siblings = adj_ast(graph, parent_id)
    current_idx = siblings.index(n_id)
    if current_idx + 1 >= len(siblings):
        return None
    next_sibling = siblings[current_idx + 1]
    if graph.nodes[next_sibling].get("label_type") != "selector":
        return None
    if not match_ast_d(graph, next_sibling, "argument_part"):
        return None
    return next_sibling


def _resolve_child_and_selector(
    graph: Graph,
    id_name: NId | None,
) -> tuple[NId | None, str | None]:
    if not id_name:
        return None, None
    child_id = match_ast_d(graph, id_name, "identifier")
    if child_id:
        return child_id, node_to_str(graph, child_id)
    return None, None


def _try_build_from_next_sibling(
    graph: Graph,
    args: SyntaxGraphArgs,
    selector_name: str,
    child_id: NId | None,
    parent_id: NId,
) -> NId | None:
    next_sibling = _get_next_arg_part_sibling(graph, args.n_id, parent_id)
    if not next_sibling:
        return None
    next_arg_id = match_ast_d(graph, next_sibling, "argument_part")
    if not next_arg_id:
        return None
    next_al_id = match_ast_d(graph, next_arg_id, "arguments")
    if not next_al_id:
        return None
    children: DirectChildren = {"arguments_id": next_al_id}
    if child_id:
        children["expression_id"] = child_id
    method_inv_id = build_method_invocation_node(args, selector_name, children)
    graph.nodes[next_sibling]["_method_invocation_id"] = method_inv_id
    return method_inv_id


def _build_final_node(
    args: SyntaxGraphArgs,
    selector_name: str | None,
    child_id: NId | None,
) -> NId:
    graph = args.ast_graph
    arg_id = match_ast_d(graph, args.n_id, "argument_part")
    if arg_id and (al_id := match_ast_d(graph, arg_id, "arguments")):
        if selector_name:
            children: DirectChildren = {"arguments_id": al_id}
            if child_id:
                children["expression_id"] = child_id
            return build_method_invocation_node(args, selector_name, children)
        return args.generic(args.fork_n_id(al_id))
    return build_selector_node(args, selector_name, None)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    if already_built := graph.nodes[args.n_id].get("_method_invocation_id"):
        return NId(already_built)

    id_name = match_ast_d(graph, args.n_id, "unconditional_assignable_selector") or match_ast_d(
        graph,
        args.n_id,
        "conditional_assignable_selector",
    )
    parent = pred_ast(graph, args.n_id)
    parent_type = graph.nodes[parent[0]]["label_type"]
    if parent_type in {"return_statement", "if_statement"} and id_name:
        pred_id = match_ast_d(graph, parent[0], "identifier")
        arg_part = match_ast_d(graph, parent[0], "argument_part", 2)
        if built_node := _reader_invocation_expression(graph, args, arg_part, id_name, pred_id):
            return built_node

    child_id, selector_name = _resolve_child_and_selector(graph, id_name)

    if selector_name and (
        method_inv_id := _try_build_from_next_sibling(
            graph, args, selector_name, child_id, parent[0]
        )
    ):
        return method_inv_id

    return _build_final_node(args, selector_name, child_id)



from blends.models import NId
from blends.syntax.models import SyntaxGraphArgs


def build_lambda_function_type_node(
    args: SyntaxGraphArgs,
    param_types: list[NId],
    return_type: NId,
) -> NId:
    # Add the lambda_function_type node to the graph
    args.syntax_graph.add_node(
        args.n_id,
        label_type="LambdaFunctionType",
    )

    # Add parameter types as children
    for param_type in param_types:
        param_id = args.generic(args.fork_n_id(param_type))
        args.syntax_graph.add_edge(
            args.n_id,
            param_id,
            label_ast="AST",
        )

    # Add return type as a child
    return_type_id = args.generic(args.fork_n_id(return_type))
    args.syntax_graph.add_edge(
        args.n_id,
        return_type_id,
        label_ast="AST",
    )

    return args.n_id



import re

from blends.models import (
    NId,
)
from blends.syntax.builders.symbol_lookup import build_symbol_lookup_node
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    label_text = args.ast_graph.nodes[args.n_id].get("label_text")
    if not label_text:
        label_text = node_to_str(args.ast_graph, args.n_id)
    symbol = re.match(r"^([a-zA-Z_][a-zA-Z0-9_]*)", label_text)
    symbol_str = symbol.group(1) if symbol else ""

    return build_symbol_lookup_node(args, symbol_str)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.return_statement import (
    build_return_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs = match_ast(graph, args.n_id, "return")
    expected_children_for_return_with_value = 2
    if len(childs) == expected_children_for_return_with_value and (stmt_id := childs["__0__"]):
        if (
            graph.nodes[stmt_id]["label_type"] == "expression_list"
            and (expr_childs := adj_ast(graph, stmt_id))
            and len(expr_childs) == 1
        ):
            return build_return_node(args, expr_childs[0])

        return build_return_node(args, stmt_id)

    return build_return_node(args, None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.array_node import (
    build_array_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes

    c_ids: set[NId] = set()

    for _id in match_ast_group_d(graph, args.n_id, "array_element_initializer"):
        element_c_ids = adj_ast(graph, _id)

        for c_id in element_c_ids:
            if nodes[c_id].get("label_type") == "=>":
                c_ids.add(c_id)
                break
        else:
            c_ids.update(element_c_ids)

    return build_array_node(args, iter(c_ids))



from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    args.evaluation[args.n_id] = False

    value_id = args.graph.nodes[args.n_id].get("value_id")
    if value_id is not None:
        args.evaluation[args.n_id] = generic(args.fork_n_id(value_id)).danger

    if args.method_evaluators and (
        method_evaluator := args.method_evaluators.get("named_argument")
    ):
        args.evaluation[args.n_id] = method_evaluator(args).danger
    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from enum import Enum
from typing import NoReturn


class PartialPathResolutionErrorCode(str, Enum):
    INCORRECT_SOURCE_NODE = "IncorrectSourceNode"
    SCOPE_STACK_UNSATISFIED = "ScopeStackUnsatisfied"
    SYMBOL_STACK_UNSATISFIED = "SymbolStackUnsatisfied"
    INCORRECT_POPPED_SYMBOL = "IncorrectPoppedSymbol"
    EMPTY_SCOPE_STACK = "EmptyScopeStack"
    EMPTY_SYMBOL_STACK = "EmptySymbolStack"
    MISSING_ATTACHED_SCOPE_LIST = "MissingAttachedScopeList"
    UNEXPECTED_ATTACHED_SCOPE_LIST = "UnexpectedAttachedScopeList"
    UNKNOWN_ATTACHED_SCOPE = "UnknownAttachedScope"
    UNBOUND_SYMBOL_STACK_VARIABLE = "UnboundSymbolStackVariable"
    UNBOUND_SCOPE_STACK_VARIABLE = "UnboundScopeStackVariable"


class PartialPathResolutionError(Exception):
    @property
    def code(self) -> "PartialPathResolutionErrorCode":
        first_arg = (
            self.args[0] if self.args else PartialPathResolutionErrorCode.SYMBOL_STACK_UNSATISFIED
        )
        return (
            first_arg
            if isinstance(first_arg, PartialPathResolutionErrorCode)
            else PartialPathResolutionErrorCode.SYMBOL_STACK_UNSATISFIED
        )


def _raise_scope_stack_unsatisfied() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.SCOPE_STACK_UNSATISFIED)


def _raise_symbol_stack_unsatisfied() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.SYMBOL_STACK_UNSATISFIED)


def _raise_incorrect_source_node() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.INCORRECT_SOURCE_NODE)


def _raise_incorrect_popped_symbol() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.INCORRECT_POPPED_SYMBOL)


def _raise_empty_scope_stack() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.EMPTY_SCOPE_STACK)


def _raise_empty_symbol_stack() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.EMPTY_SYMBOL_STACK)


def _raise_missing_attached_scope_list() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.MISSING_ATTACHED_SCOPE_LIST)


def _raise_unexpected_attached_scope_list() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.UNEXPECTED_ATTACHED_SCOPE_LIST)


def _raise_unknown_attached_scope() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.UNKNOWN_ATTACHED_SCOPE)


def _raise_unbound_symbol_stack_variable() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.UNBOUND_SYMBOL_STACK_VARIABLE)


def _raise_unbound_scope_stack_variable() -> NoReturn:
    raise PartialPathResolutionError(PartialPathResolutionErrorCode.UNBOUND_SCOPE_STACK_VARIABLE)



from blends.models import (
    NId,
)
from blends.syntax.builders.for_statement import (
    build_for_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    body_id = n_attrs["label_field_body"]
    var_node = n_attrs["label_field_pattern"]
    condition_node = n_attrs["label_field_value"]

    return build_for_statement_node(args, var_node, condition_node, None, body_id)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
    match_ast_group_d,
)
from blends.syntax.builders.try_statement import (
    build_try_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    block_node = args.ast_graph.nodes[args.n_id]["label_field_body"]
    childs = match_ast(
        args.ast_graph,
        args.n_id,
        "catch_clause",
        "finally_clause",
    )
    catch_blocks = match_ast_group_d(args.ast_graph, args.n_id, "catch_clause")
    try_block = childs.get("finally_clause")

    return build_try_statement_node(args, block_node, catch_blocks, try_block, None)



from __future__ import annotations

from mf_transport import (
    Request,
    Response,
    ConnectionPool,
    TransportError,
    send_with_retry,
    SENTINEL,
)
from mf_codec import Frame, Encoder, Decoder, CodecError, encode_batch
from mf_core import RetryPolicy, MAX_RETRIES


class ProcessorError(TransportError):
    stage: str = ""
    codec_error: CodecError = CodecError()

    def describe(self: "ProcessorError") -> str:
        base: str = TransportError.describe(self)
        return base + " stage=" + self.stage


class ProcessorConfig:
    max_batch_size: int = 32
    max_retries: int = MAX_RETRIES
    retry: RetryPolicy = RetryPolicy()
    pool: ConnectionPool = ConnectionPool()
    encoder: Encoder = Encoder()
    decoder: Decoder = Decoder()


class ProcessingResult:
    success: bool = False
    frames_encoded: int = 0
    payload_decoded: int = 0
    response: Response = Response()
    error: ProcessorError = ProcessorError()

    def is_ok(self: "ProcessingResult") -> bool:
        return self.success and self.response.is_success()

    def total_data(self: "ProcessingResult") -> int:
        return self.frames_encoded + self.payload_decoded

    def has_error(self: "ProcessingResult") -> bool:
        return not self.success


def process_single(config: ProcessorConfig, payload_size: int) -> ProcessingResult:
    result: ProcessingResult = ProcessingResult()
    req: Request = Request()
    req.retry = config.retry
    req.payload_size = payload_size
    frames: int = config.encoder.encode(payload_size)
    if frames == SENTINEL:
        result.error.stage = "encode"
        return result
    result.frames_encoded = frames
    resp: Response = send_with_retry(config.pool, req)
    result.response = resp
    if resp.is_success():
        decoded: int = config.decoder.decode(frames)
        result.payload_decoded = decoded
        result.success = True
    else:
        result.error.stage = "transport"
        result.error.transport_code = resp.status
    return result


def process_batch(config: ProcessorConfig, sizes: list) -> int:
    total_frames: int = encode_batch(config.encoder, sizes)
    success_count: int = 0
    idx: int = 0
    while idx < len(sizes):
        result: ProcessingResult = process_single(config, sizes[idx])
        if result.is_ok():
            success_count = success_count + 1
        idx = idx + 1
    return success_count


class Processor:
    config: ProcessorConfig = ProcessorConfig()
    total_processed: int = 0
    total_failed: int = 0
    total_frames: int = 0
    max_retries: int = MAX_RETRIES

    def run(self: "Processor", payload_size: int) -> ProcessingResult:
        result: ProcessingResult = process_single(self.config, payload_size)
        if result.is_ok():
            self.total_processed = self.total_processed + 1
            self.total_frames = self.total_frames + result.frames_encoded
        else:
            self.total_failed = self.total_failed + 1
        return result

    def run_batch(self: "Processor", sizes: list) -> int:
        count: int = process_batch(self.config, sizes)
        self.total_processed = self.total_processed + count
        self.total_failed = self.total_failed + len(sizes) - count
        return count

    def success_rate(self: "Processor") -> int:
        total: int = self.total_processed + self.total_failed
        if total == 0:
            return 100
        return self.total_processed * 100 // total

    def reset(self: "Processor") -> None:
        self.total_processed = 0
        self.total_failed = 0
        self.total_frames = 0


GLOBAL_PROCESSOR: Processor = Processor()
DEFAULT_SENTINEL: int = SENTINEL
MAX_RETRY_LIMIT: int = MAX_RETRIES



import asyncio
import json
import os as operatingsys
from collections.abc import (
    Callable,
)
from itertools import (
    accumulate,
)
from itertools import (
    chain as join_chain,
)

FIELDS_BY_LANGUAGE: dict[str, tuple[str, ...]] = {}


class InsufficientAmount(Exception):
    pass


class Wallet(dict[str, int]):
    user_id: str
    balance: int
    withdrawals: int

    def __init__(self, user_id: str, initial_amount: int = 0) -> None:
        self.user_id = user_id
        self.balance = initial_amount
        self.withdrawals = 0

    def spend_cash(self, amount: int) -> None:
        if self.balance < amount:
            raise InsufficientAmount(f"Not enough available to spend {amount}")
        self.balance -= amount

    def add_cash(self, amount: int) -> None:
        if self.balance > 1000000000:
            raise InsufficientAmount(f"Careful {operatingsys.environ['MILLIONAIRE']}")
        self.balance += amount


def main() -> None:
    number = 100
    modulus = number % 2
    if modulus == 0:
        print("The number is Even.")
    else:
        print("The number is Odd.")

    sum_var = 0
    count = 10
    for i in range(count):
        sum_var = sum_var + i
    avg = sum_var / count
    print("The average is: ", avg)

    count = 0
    mult_var = 1
    while count < 10:
        count = count + 1
        mult_var *= number


def comprehensions() -> None:
    input_list = [1, 2, 3, 4, 4, 5, 6, 7, 7]
    list_using_comp = [var for var in input_list if var % 2 == 0]
    print(list_using_comp)
    set_using_comp = {var for var in input_list if var % 2 == 0}
    print(set_using_comp)
    dict_using_comp = {var: var**3 for var in input_list if var % 2 != 0}
    print(dict_using_comp)

    state = ["Gujarat", "Maharashtra", "Rajasthan"]
    capital = ["Gandhinagar", "Mumbai", "Jaipur"]
    output_dict = {}
    for key, value in zip(state, capital, strict=False):
        output_dict[key] = value

    try:
        my_numbers = [1, 2, 3]
        accumulate(my_numbers)
        join_chain(my_numbers)
        print(my_numbers[10])
    except IndexError as err:
        if str(err) == "Error":
            print("Error")
    finally:
        # some code that should always execute
        print("This code will always execute!")


def get_artifact(env_var: str) -> str:
    if value := env_var:
        return value
    raise ValueError(f"Expected environment variable: {env_var}")


TREE_SITTER_PARSERS = get_artifact("SKIMS_TREE_SITTER_PARSERS")


def get_fields_by_language() -> None:
    for lang in ["spanish", "english"]:
        if len(lang) < 3:
            continue
        if not lang.startswith("s"):
            break
        path: str = operatingsys.path.join(TREE_SITTER_PARSERS, f"{lang}.json")
        with open(path, encoding="utf-8") as file:
            FIELDS_BY_LANGUAGE[lang] = json.load(file)


def capital_case(country_name: str) -> str:
    return country_name.capitalize()


def test_capital_case() -> None:
    assert capital_case("semaphore") == "Semaphore"


def my_decorator(func: Callable) -> Callable:
    def wrapper() -> None:
        print("Before the function is called.")
        func()
        print("After the function is called.")

    return wrapper


@my_decorator
def say_hello() -> None:
    print("Hello!")


async def my_async_function() -> None:
    print("Start")
    await asyncio.sleep(1)  # Wait for one second
    print("End")


def all_parameters(
    *args: int,
    positional_arg: str = "hello",
    default_arg: str = "world",
    keyword_only_arg: int,
    **kwargs: str,
) -> None:
    print("positional_arg:", positional_arg)
    print("default_arg:", default_arg)
    print("args:", args)
    print("keyword_only_arg:", keyword_only_arg)
    print("kwargs:", kwargs)


all_parameters(3, 4, keyword_only_arg=42, another_key="hi")

dino_kind: str = "herbivore"
a_dinosaurs_name: str = "T-Rex" if dino_kind == "carnivore" else "Triceratops"



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    var_name = "unnamed"
    childs = match_ast(args.ast_graph, args.n_id, "namespace")
    if var_id := childs.get("__0__"):
        var_name = node_to_str(args.ast_graph, var_id)

    return build_variable_declaration_node(args, var_name, None, None)
